""" Tests for spatial occupancy"""
from opexebo.general import shuffle as func

import numpy as np
import pytest

print("=== tests_general_shuffle ===")

###############################################################################
################                MAIN TESTS
###############################################################################


def test_output_structure():
    """Verify that the function correctly returns an array where each row is
    a time-shifted copy of the input, with num_shuffle rows"""

    times = np.random.randint(0, 100, size=1000)
    offset_lim = 13
    iterations = 274
    tr = np.arange(111)
    output, _ = func(times, offset_lim, iterations, tracking_range=tr)

    # output.shape = rows, cols
    assert output.shape[0] == iterations
    assert output.shape[1] == times.size
    print("test_output_structure() passed")


def WIP_test_consistent_offset():
    '''All the extra faff here is to deal with the fact that since times does
    not necessarily fill the entire tracking range. Information can be "lost"
    to np.diff because it's not circular (i.e. it doesn't compare last to first)
    Therefore, we manually add this value in at the end "miss"'''
    tr = np.arange(173)
    times = np.random.randint(0, 100, size=10)
    times = np.sort(times)
    offset_lim = 13
    iterations = 2

    output, inc = func(times, offset_lim, iterations)
    miss = (np.min(times) - np.min(tr)) + (np.max(tr) - np.max(times))
    inc = np.atleast_2d(inc + miss).transpose()
    output = np.hstack((output, inc))
    times = np.append(times, miss)

    true_d = np.diff(times)
    # true_d = np.sort(true_d).astype(int)
    d = np.diff(output, axis=1)

    # The order of each row of d **should** be preserved circularly, but we can't
    # just compare like-for-like
    # The simplest, but not best, way of comparison is to sort both
    print(times)
    print(true_d)
    for r in range(iterations):
        row = d[r].astype(int)
        print(row)
    print("test_consistent_offset() passed")


def test_stupid_values():
    times = np.random.randint(0, 100, size=1000)
    offset_lim = 13
    iterations = 274
    tr = np.arange(111)
    with pytest.raises(ValueError):
        stupid_times = np.random.randint(0, 100, size=(100, 100))  # wrong shape
        func(stupid_times, offset_lim, iterations, tracking_range=tr)
    with pytest.raises(ValueError):
        stupid_times = times.copy().astype(float)
        stupid_times[813] = np.nan  # Contains NaN
        func(stupid_times, offset_lim, iterations, tracking_range=tr)
    with pytest.raises(ValueError):
        stupid_offset = 120
        func(times, stupid_offset, iterations, tracking_range=tr)
    print("test_stupid_values() passed")


# if __name__ =='__main__':
#    test_output_structure()
#    test_consistent_offset()
#    test_stupid_values()



""" Tests for autocorrelogram"""
import numpy as np

import opexebo.tests as th
from opexebo.analysis import autocorrelation as func

print("=== tests_analysis_autocorrelation ===")


###############################################################################
################                MAIN TESTS
###############################################################################


def test_random_input():
    firing_map = np.random.rand(80, 80)
    acorr = func(firing_map)
    return acorr


def test_perfect_artificial_grid():
    firing_map = th.generate_2d_map(
        "rect",
        1,
        x=80,
        y=80,
        coverage=0.95,
        fields=th.generate_hexagonal_grid_fields_dict(),
    )
    acorr = func(firing_map)
    return acorr



import numpy as np

def spatial_cross_correlation(arr_0, arr_1, **kwargs):
    '''
    Calculate the Pearson cross correlation between two arrays. 
    
    Arrays may be either 1d or 2d. If MaskedArrays are supplied, all locations
    that are masked are excluded from the correlation. Invalid numbers (NaN) 
    are also excluded from the correlation.

    If 2d arrays are provided, then calculates a 1d array of row(or column)-wise
    correlations, controlled by the `row_major` keyword.

    Parameters
    ----------
    arr_0: np.ndarray
        1d or 2d array. Must either contain no NaNs, or be an np.ma.MaskedArray
        with NaN values masked. 
    
    arr_1: np.ndarray
        2nd array to correlate with. Must have the same dimensions as `arr_0`
    
    Other Parameters
    ----------------
    row_major: bool, optional
        If the arrays are 2D, process first by row, or first by column.
        Default `True`

    Returns
    -------
    output_single: float
        Pearson correlation coefficient between entire arrays. 2d input arrays
        are flattened to 1d to calculate
    output_array: np.ndarray
        1D array of pearson correlation coefficients, where the i'th value is
        the coefficient between the i'th rows (if row_major), or the i'th columns
        of the two arrays
        Return value is np.nan if 1d arrays are supplied

    Notes
    --------
    BNT.+analyses.spatialCrossCorrelation()

    Copyright (C) 2019 by Simon Ball
    '''
    debug = kwargs.get("debug", False)
    arr_0 = _check_single_array(arr_0, 0, **kwargs)
    arr_1 = _check_single_array(arr_1, 1, **kwargs)        
    
    if arr_0.shape != arr_1.shape:
        raise ValueError(f"Your arrays are different shapes: {arr_0.shape} and"\
                         " {arr_0.shape}. You must provide the same shape arrays")

    if arr_0.ndim not in (1, 2):
        raise ValueError(f"Your arrays have the wrong number of dimensions. Only"\
                         "1d and 2d arrays are supported. You provided {arr_0.ndim}"\
                         "dimensions")

    is_2d = arr_0.ndim==2

    joint_mask = np.logical_or(arr_0.mask, arr_1.mask)
    good = np.logical_not(joint_mask)
    if debug:
        print(f"is_2d: {is_2d}")
        print(f"{np.sum(joint_mask)} values excluded")

    output_single = np.corrcoef(arr_0.data[good].flatten(), arr_1.data[good].flatten())[0,1]
    output_array = np.nan

    if is_2d:
        row_major = kwargs.get("row_major", True)
        y_bins, x_bins = arr_0.shape
        if row_major:
            iterator = y_bins
        else:
            iterator = x_bins
        output_array = np.zeros(iterator)

        for i in range(iterator):
            if row_major:
                left = arr_0[i, :][good[i, :]]
                right = arr_1[i, :][good[i, :]]
            else:
                left = arr_0[:, i][good[:, i]]
                right = arr_1[:, i][good[:, i]]
            output_array[i] = np.corrcoef(left, right)[0,1]

    return output_single, output_array


def _check_single_array(arr, i, **kwargs):
    '''input checking on a single array'''
    if isinstance(arr, np.ndarray):
        arr = np.ma.masked_invalid(arr)
    elif not isinstance(arr, np.ma.MaskedArray):
        raise ValueError(f"arr_{i} is type {type(arr)}. You must provide a Numpy"\
                         "array or MaskedArray")
    return arr


""" Provides a function for Gaussian smoothing """
# Hack to speed up the borked Astropy configuration search
import os
import pathlib
os.environ["HOMESHARE"] = str(pathlib.Path.home())

import numpy as np
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    # Suppress the ConfigurationMissingWarning that Astropy triggers
    from astropy.convolution import convolve, Gaussian2DKernel, Gaussian1DKernel

import opexebo.defaults as default
#http://docs.astropy.org/en/stable/convolution/index.html
#
#Astropy appears to have an import problem - it routinely takes about 3 minutes
#to import, apparently because it is looking for a file with a stupidly long
#timeout. It typically results in a warning:
#        ConfigurationMissingWarning : Configuration defaults will be used due
#           to FileNotFoundError:2 on None
#
#The root cause seems to be that Astropy expects to find a configuration file
#in a very specific location - in thise case, a network location
#       (\\home.ansatt.ntnu.no\.astropy)
#Which doesn't exist. But the timeout for checking the network is very long
#
# A tempoary workaround is to create the environment variables
#    XDG_CACHE_HOME
#    XDG_CONFIG_HOME
# pointing to local locations that do exist. Since we only use a very minor
# part of Astropy here, the fact that the config file may be ephemeral is not a
# problem. See here for discussion of these variables
# https://github.com/astropy/astropy/issues/6511


def smooth(data, sigma, **kwargs):
    '''Smooth provided data with a Gaussian kernel

    The smoothing is done with a routine from the astronomical package astropy
    Like scipy.ndimage.gaussian_filter, this does not handle MaskedArrays - but
    it handles NaNs much better. Specifically, astropy.convolution.convolve
    replaces NaN values with an interpolation across the void region.

    Therefore, to handle masked arrays, the data at masked positions *is
    replaced by np.nan* prior to smoothing, and thus avoids influencing nearby,
    unmasked cells. The masked cells are then returned to their original values
    prior to return.

    The package is discussed at
    http://docs.astropy.org/en/stable/convolution/index.html

    Parameters
    ----------
    data: np.ndarray or np.ma.MaskedArray
        Data that will be smoothed
    sigma: float
        Standard deviations for Gaussian kernel in units of pixels/bins
    mask_fill: float, optional
        The value that masked locations should be treated as
        This can either be provided as an absolute number (e.g. 0), or nan
        If nan, then each masked location will get a value by interpolating
        from nearby cells. This will only apply if the input is a
        MaskedArray to start with. If the input is a standard np.ndarray,
        then no values will be substituted, even if there are nans present.
    circular: bool, optional
        If True, then smoothing at the edge of the array will be handled in
        a circular manner, i.e. the value to the left of data[0] will be
        data[-1]. If False, the edge will be handled by padding with values
        equal to the boundary value. Default False

    Returns
    -------
    smoothed_data: np.ndarray or np.ma.MaskedArray
        Smoothed data

    Notes
    --------
    BNT.+general.smooth
    http://docs.astropy.org/en/stable/convolution/index.html
    https://github.com/astropy/astropy/issues/6511

    Copyright (C) 2019 by Simon Ball
    '''
    d = data.ndim
    if  d == 2:
        kernel = Gaussian2DKernel(x_stddev=sigma)
    elif d == 1:
        kernel = Gaussian1DKernel(stddev=sigma)
    else:
        raise NotImplementedError("This function currently supports smoothing"\
                f" 1D, 2D data. You have provided {d} dimensional data")

    mask_fill = kwargs.get('mask_fill', default.mask_fill)
    circular = kwargs.get("circular", False)
    if not isinstance(circular, bool):
        raise ValueError("You must provide a boolean (True/False) value for"\
                         f" keyword 'circular'. You provided {circular}, which"\
                         f" is type {type(circular)}")

    working_data = data.copy()
    if type(data) == np.ma.MaskedArray:
        working_data[data.mask] = mask_fill

    width = int(4*sigma)

    if circular:
        smoothed_data = convolve(working_data, kernel, boundary="wrap")
        # Don't bother with padding. Use the values from the other end of the 
        # array, i.e. imagine the array wrapped around a cylinder
    elif not circular:
        working_data = np.pad(working_data, pad_width=width, mode='symmetric')
        # pad the outer boundary to depth "width
        # The padding values are based on reflecting at the border
        # mode='symmetrical' results in
        # [0, 1, 2, 3, 4] -> [1,0  ,0,1,2,3,4,  4,3]
        # mode='reflect' results in
        # [0, 1, 2, 3, 4] -> [2,1  ,0,1,2,3,4,  3,2]
        # i.e. changing whether the reflection axis is outside the original data
        # or overlaid on the outermost row
    
        smoothed_data = convolve(working_data, kernel, boundary='extend')
        # Because of the padding, the boundary mode isn't really relevant
        # By choosing a large width, the edge effects arising from this additional
        # padding (boundary='extend') is minimised
    
        if d == 2:
            smoothed_data = smoothed_data[width:-width, width:-width]
        elif d == 1:
            smoothed_data = smoothed_data[width:-width]
        else: # This condition should never happen, due to checking above
            raise NotImplementedError("This function currently supports smoothing"\
                    f" 1D, 2D data. You have provided {d} dimensional data")
        # We have to get rid of the padding that we previously added, and the only
        # way to do that is slicing, which is NOT dimensional-agnostic
        # There may be a more elegant solution than if/else, but this will do now

    if type(data) == np.ma.MaskedArray:
        smoothed_data = np.ma.masked_where(data.mask, smoothed_data)
        smoothed_data.data[data.mask] = data.data[data.mask]
    
    assert smoothed_data.shape == data.shape, "Output array is a different shape to input array"

    return smoothed_data



""" Tests for grid score"""
import opexebo.tests as th
import opexebo
from opexebo.analysis import grid_score as func

print("=== tests_analysis_grid_score ===")


###############################################################################
################                MAIN TESTS
###############################################################################


def test_perfect_grid_cell():
    firing_map = th.generate_2d_map(
        "rect",
        1,
        x=80,
        y=80,
        coverage=0.95,
        fields=th.generate_hexagonal_grid_fields_dict(),
    )
    acorr = opexebo.analysis.autocorrelation(firing_map)
    gs, stats = func(acorr, debug=True, search_method="default")
    assert stats["grid_ellipse_aspect_ratio"] > 0.985
    assert stats["grid_ellipse_aspect_ratio"] < 1.15


# if __name__ == '__main__':
#    test_perfect_grid_cell()



"""Provide function for calculating coherence of a Rate Map"""
# Hack to speed up the borked Astropy configuration search
import os
import pathlib
os.environ["HOMESHARE"] = str(pathlib.Path.home())

import numpy as np
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    # Suppress the ConfigurationMissingWarning that Astropy triggers
    from astropy.convolution import convolve


def rate_map_coherence(rate_map_raw):
    '''
    Calculate coherence of a rate map

    Coherence is calculated based on RU Muller, JL Kubie "The firing of hippocampal place
    cells predicts the future position of freely moving rats", Journal of Neuroscience, 1 December 1989,
    9(12):4101-4110. The paper doesn't provide information about how to deal with border values
    which do not have 8 well-defined neighbours. This function uses zero-padding technique.

    Parameters
    ----------
    rate_map_raw: np.ma.MaskedArray
        Non-smoothed rate map: n x m array where cell value is the firing rate,
        masked at locations with low occupancy

    Returns
    -------
    coherence: float
        see relevant literature (above)

    Notes
    -----
    BNT.+analyses.coherence(map)
        
    Copyright (C) 2019 by Simon Ball
    '''

    kernel = np.array([[0.125, 0.125, 0.125],
                      [0.125, 0,     0.125],
                      [0.125, 0.125, 0.125]])

    avg_map = convolve(rate_map_raw, kernel, 'fill', fill_value=0)
    avg_map = avg_map.ravel()
    avg_map = np.nan_to_num(avg_map)

    rmap = np.copy(rate_map_raw)
    rmap = rmap.ravel()
    rmap = np.nan_to_num(rmap)

    coherence = np.corrcoef(avg_map, rmap)[0,1]

    return coherence


"""Provide function for calculating a Border Score"""

import numpy as np
from scipy.ndimage import distance_transform_cdt
try:
    import cv2
    HAS_CV2 = True
except:
    HAS_CV2 = False


from opexebo.analysis import border_coverage
from opexebo.general import validate_keyword_arena_shape, circular_mask
import opexebo.defaults as default


def border_score(rate_map, fields_map, fields, arena_shape, **kwargs):
    '''
    Calculate border score for a firing map
    
    STATUS : EXPERIMENTAL
    
    TODO: Update to account for the use of Vadim's placefield code
    TODO: Handle circular arenas (How?)
    

    Calculates a border score for a firing rate map according to the article 
    "Representation of Geometric Borders in the Entorhinal Cortex" by Solstad 
    et. al. (Science 2008).
    
    Border score is either -1 (no firing fields provided) or in the range [-1, 1].
    The score reflects both the width of a field (what fraction of a single wall
    it touches), and the depth of a field (how far away from the field it extends)
    The highest score is returned for a field with maximum width and infinitesimal
    depth.
    
    Consequently, a border score of 1 can only **ever** be returned given an 
    infinite resolution. A perfect field in a typical 40x40 bin map has a maximum
    value of around 0.91. (limited by the width of the binning compared to the
    arena extent)
    
    The score is only evaluated for the single firing field that has the greatest
    wall coverage. All other fields are ignored. 
    
    Parameters
    ----------
    rate_map:   np.ma.MaskedArray
        rate map: N x M array where each cell value is the firing rate of the cell
    fields_map:   np.ma.MaskedArray
        Integer array of labelled fields. Each cell value is a positive integer. 
        Cells that are members of field have the value corresponding to field_id 
        (non-zero positive  unique integers. Not necessarily contiguous 
        (e.g. 1, 2, 3, 5)). Cells that are not members of fields have value zero
    fields:  list of  dict
        List of dictionaries of firing fields. 
        Each dictionary must, at least, contain the keyword "field_map", yielding a 
        binary map of that field within the overall arena
    arena_shape: {"square", "rect", "circle", "line"}
        Rectangular and square are equivalent. Elliptical or n!=4 polygons
        not currently supported.
    search_width: int, optional
        rate_map and fields_map have masked values, which may occur within the
        region of border pixels. To mitigate this, we check rows/columns within
        search_width pixels of the border Default `8`
    walls: str or list of tuple, optional
        Definition of which walls to consider for border coverage. Behaviour is
        different for different arena shapes
        * Rectangular arenas
          type str. The four walls are referred to as `T`, `B`, `R`, `L` for the
          standard cardinals. Case insensitive. You may include or exclude any
        * Circular arenas
          Walls should be specified as a list of inclusive angular start-stop pairs. Angles
          given in degrees clockwise from top. Both `0` and `360` are permissible
          values. Start-stop may wrap through zero and may be overlapping with
          other pairs. Negative angles may be used to express counter-clockwise
          rotation, and will be automatically wrapped into the [0, 360] range.
          Example: [(0, 15), (180, 270), (315, 45), (30, 60)]
          Walls may ALTERNATIVELY be expressed as a string, equivalent to rectangular
          arenas. In this case, walls will be interpreted as follows
              * `T` - (315, 45)
              * `B` - (135, 225)
              * `L` - (45, 135)
              * `R` - (225, 315)
        In either case, defaults `TRBL`. Converting this (or other string) for
        circluar arrays is handled internally

    Returns
    -------
    score: float
        Border score in the range [-1, 1].
        A value of -1 is given when no fields are provided
    coverage: float
        Border coverage in the range [0, 1]
        The coverage is given for ONE border. Coverage for a specific wall is
        only available if a single wall is searched

    See Also
    --------
    opexebo.analysis.placefield
    opexebo.analysis.bordercoverage

    Notes
    -----
    * BNT.+analyses.placefield
    * BNT.+analyses.borderScore
    * BNT.+analyses.borderCoverage
        
    Copyright (C) 2019 by Simon Ball
    '''
    arena_shape = validate_keyword_arena_shape(arena_shape)

    # Check that fields exist
    if np.ma.max(fields_map) == 0:
        # No fields exist
        return -1, 0
    
    # 1 or more fields exist
    fields_map_unlabelled = np.copy(fields_map)
    fields_map_unlabelled[fields_map> 1 ] = 1

    coverage = border_coverage(fields, arena_shape, **kwargs)

    fields_rate_map = fields_map_unlabelled * rate_map
    fields_rate_map = np.ma.masked_invalid(fields_rate_map)

    wfd = _weighted_firing_distance(fields_rate_map, arena_shape)

    score = (coverage - wfd)/(coverage + wfd)
    return score, coverage



###############################################################################
####            Border Score helper functions
###############################################################################


def _weighted_firing_distance(rmap, arena_shape):
    '''
    Calculate the weighted firing distance
    
    A perfect border cell is one that fires only (and always) when the animal is
    imemdiately adjacent to a border. This function weights the firing rate
    reported in the ratemap by its proximity to the border.

    Parameters
    ----------
    rmap: np.ma.MaskedArray
        Rate map of cells inside fields. NxM array where cells inside fields
        give the firing rate, and cells outside fields are zero
    Returns
    -------
    wfd: float
        Weighted firing distance. 
    '''
    # Check that the provided array is properly masked
    if type(rmap != np.ma.MaskedArray):
        rmap = np.ma.masked_invalid(rmap)

    # Normalise firing map by the sum of the firing map, such that it resembles a PDF
    pdf = rmap / np.ma.sum(rmap)

    # Create an array, same size as rmap, where the cell value is the distance (in bins) from the edge of the array:
    x = np.ones(rmap.shape) # define area same size as ratemap
    x = np.pad(x, 1, mode='constant') # Pad outside with zeros, we calculate distance from nearest zero
    dist = distance_transform_cdt(x, metric='taxicab') # taxicab metric: distance along axes, not diagonal distance. 
    #wfd = np.ma.sum( np.ma.dot(dist[1:-1, 1:-1], rmap) )
    wfd = np.ma.sum( dist[1:-1, 1:-1] * pdf )

    # Normalise by half of the smallest arena dimensions
    wfd = 2 * wfd / np.min(rmap.shape)
    return wfd



def _identify_border(rmap, arena_shape, debug=False):
    '''
    Identify the border based on the arena_shape. This is needed to cope with
    circular arenas
    
    Parameters
    ----------
    rmap: np.ndarray or np.ma.MaskedArray
        Rate map of the unit
    arena_shape: {"square", "rect", "circle", "line"}
        Rectangular and square are equivalent. Elliptical or n!=4 polygons
        not currently supported.
    
    Returns
    -------
    dict
        mask: np.ndarray
            Boolean array. `True` inside arena, `False` outside
        distance: np.ndarray, optional
            [Circular arenas only] Array of distances from the border
        angle: np.ndarray, optional
            [Circular arenas only] Array of angles from the centre. 
    '''
    out = {}
    if arena_shape in default.shapes_square:
        mask = np.ones(rmap.shape, dtype=bool)
        out["mask"] = mask
    elif arena_shape in default.shapes_circle:        
        if not HAS_CV2:
            raise ImportError("You must install OpenCV to use this function: `pip install opencv-python`")
        rmap_threshold = np.uint8(rmap>0)
        centres = cv2.findContours(rmap_threshold, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[-2]
        x = y = radius = 0
        for c in centres:
            (x_, y_), r_ = cv2.minEnclosingCircle(c)
            if r_ > radius:
                x = x_
                y = y_
                radius = r_
        axes = [np.arange(0, rmap.shape[i]+1, 1) for i in [0,1]]
        mask, distance, angle = circular_mask(axes, radius*2, origin=(x, y))
        out["mask"] = mask
        out["distance"] = radius - distance
        out["angle"] = angle
        if debug:
            print(f"radius (bins): {radius}")
    else:
        raise NotImplementedError("arena shape '{}' not supported".format("arena_shape"))
    return out
        


import numpy as np
from scipy import signal

import opexebo.defaults as default


def power_spectrum(values, time_stamps, sampling_frequency, **kwargs):
    '''
    Calculate the power spectrum of a time-series of data
    
    Parameters
    ----------
    values: np.ndarray
        Amplitude of time series at times `time_stamps`. Must be sampled at a
        constant frequency
    time_stamps: np.ndarray
        Time stamps of time-series. Must be the same length as `values`. 
        Assumed to be in [seconds]
    sampling_frequency: float
        Sampling rate. Assumed to be in [Hz]
    method: str, optional, {"welch", "fft"}
        Method for calculating a power spectrum. Accepted values are:
        ["welch", "fft"]. Defaults to "welch"
    scale: str, optiona, {"linear", "log"}
        Should the power spectrum be returned in a linear form 
        propto v**2/sqrt(Hz)) ("linear"); 
        or decibel scale (dB/sqrt(Hz)) ("log")
        Defaults to "linear"
    resolution: float
        Desired output frequency resolution. Applicable only to Welch's 
        method. Due to the averaging effect of Welch's method, this sets
        the "effective" resolution, which is NOT the same as the minimum
        difference in `frequencies`
        Default 1 [Hz]. Values lower than 1/(time_series_duration) are 
        meaningless
    
    Returns
    -------
    frequencies: np.ndarray
        Discrete Fourier Transform sample frequencies
    power_spectrum: np.ndarray
        Power spectral density at that sample frequency
    '''
    method = kwargs.get("method", default.power_spectrum_method).lower()
    scale = kwargs.get("scale", default.psd_return_scale).lower()

    if scale not in ("linear", "log"):
        raise ValueError(f"Keyword `scale` must be one of ('linear', 'log').")

    if method == "welch":
        func = _power_spectrum_welch
    elif method == "fft":
        func = _power_spectrum_fft
    else:
        raise NotImplementedError(f"Method '{method}' not implemented")

    frequencies, power_spectrum = func(values, time_stamps, sampling_frequency, **kwargs)

    if scale == "log":
        power_spectrum = 20 * np.log10(power_spectrum)
    
    return frequencies, power_spectrum




def _power_spectrum_welch(values, time_stamps, sampling_frequency, **kwargs):
    '''Use Welch's method to calculate a power spectrum for a time-series of
    data
    '''
    resolution_multiplier = 32 
    output_resolution = kwargs.get("resolution", default.psd_resolution_welch)
    
    n = sampling_frequency * resolution_multiplier / output_resolution
    
    freqs, power_spectral_density = signal.welch(x=values, fs=sampling_frequency, nperseg=n, scaling="density")
    power_spectral_density = np.sqrt(power_spectral_density)
    
    return freqs, power_spectral_density




def _power_spectrum_fft(values, time_stamps, sampling_frequency, **kwargs):
    '''Calculate the simplest, noisiest, power spectrum by taking the real
    component of the fast Fourier transform of the time-series  
    '''
    sampling_time_period = 1/sampling_frequency
    
    amplitude_spectral_density = np.abs(np.fft.rfft(values))
    power_spectral_density = np.square(amplitude_spectral_density)
    freqs = np.fft.rfftfreqs(values.size, sampling_time_period)
    
    return freqs, power_spectral_density



""" Tests for turning curve stats"""
import numpy as np
import inspect

from opexebo.analysis import tuning_curve_stats as func

print("=== tests_analysis_turning_curve_stats ===")


###############################################################################
################                MAIN TESTS
###############################################################################


def test_random_data():
    n = 1000
    data = np.random.rand(n) * 2 * np.pi
    tuning_curve = np.histogram(data, bins=int(360 / 15))[0]
    func(tuning_curve)
    print(f"{inspect.stack()[0][3]} passed")
    return True


# if __name__ == '__main__':
#    test_random_data()



import numpy as np
import opexebo.defaults as default
from skimage import measure, morphology
from scipy.ndimage import distance_transform_cdt


#import sep
# NOTE!! Due to how annoying it is to install sep, I have moved this dependency 
# into the method that uses it. This will allow users that struggle to install it
# to proceed with other methods that do not rely on it

# TODO! Decide what, if any, parameters need to move over to Defaults


def peak_search(image, **kwargs):
    """Given a 1D or 2D array, return a list of co-ordinates of the local 
    maxima or minima
    
    Multiple searching techniques are provided:
        
        * `default`: uses `skimage.morphology.get_maxima`
        * `sep`: uses the Python wrapper to the Source Extractor astronomy tool
          to identify peaks
    
    Parameters
    ----------
    image: np.ndarray
        1D or 2D array of data
    search_method : str, optional, {"default", "sep"}
    mask : np.ndarray, optional
        Array of masked locations in the image with the same dimensions.
        Locations where the mask value is True are ignored for the purpose of
        searching.
    maxima: bool, optional
        [`default` search method only] Define whether to search for maxima or
        minima in the provided array
    null_background: bool
        [`sep` search method only] Set the image background to zero for
        calculation purposes rather than attempt to calculate a background
        gradient. This should generally be True, as our images are not directly
        comparable to standard telescope output
    threshold : float, optional
        [`sep` search method only] Threshold for identifiying maxima area
    
    Returns
    -------
    peak_coords: tuple
        Co-ordinates of peaks, in the form ((x0, x1, x2...), (y0, y1, y2...))
    
    
    Notes
    --------
    Copyright (C) 2019 by Simon Ball
    """
    
    search_method = kwargs.get("search_method", default.search_method)
    get_maxima = kwargs.get("maxima", True)
    
    if search_method not in default.all_methods:
        raise ValueError("Keyword 'search_method' must be left blank or given a"\
                    " value from the following list: %s. You provided '%s'."\
                    % (default.all_methods, search_method) )
    if search_method != "default" and not get_maxima:
        raise NotImplementedError("Local minima detection is currently only"\
                            " implemented for the 'default' search method")
        
    if search_method == default.search_method:
        peak_coords = _peak_search_skimage(image, **kwargs)
    elif search_method == "sep":
        peak_coords = _peak_search_sep_wrapper(image, **kwargs)
    else:
        raise NotImplementedError("The search method you have requested (%s) is"\
                                  " not yet implemented" % search_method)
        
    return peak_coords
        
        
        
        
#########################################################
################        Helper Functions
#########################################################


def _peak_search_skimage(image, **kwargs):
    '''Default peak detection method:
        skimage.morphology.get_m**ima (either minima or maxima)
    Since skimage doesn't handle masked arrays, the masking is a bit of a bodge
    job here. The basic plan is as follows:
        * Set the area of the image covered by the mask to a value that cannot include a maxima (or minima, as appropriate)
        * Search for peak coordinates
        * Check if, after rounding, any of the peaks are outside the image dimensions
        * Check if, after rounding, any of the peaks are extremely close to the mask
    
    Since the mask is a ahrd-edged area, if there is even a slight rise just
    outside it, spurious peaks can be detected. Therefore, we automatically reject
    any peaks for a short distance outside the actual mask
    
    '''
    connectivity = 2
    get_maxima = kwargs.get("maxima", True)
    mask = kwargs.get("mask", np.zeros(image.shape, dtype=bool))
    image_copy = image.copy()
    if get_maxima:
        image_copy[mask] = np.nanmin(image_copy)
        regionalMaxMap = morphology.local_maxima(image_copy, connectivity=connectivity, allow_borders=True)
    else:
        image_copy[mask] = np.nanmax(image_copy)
        regionalMaxMap = morphology.local_minima(image_copy, connectivity=connectivity, allow_borders=True)
    labelled_max = measure.label(regionalMaxMap, connectivity=connectivity)
    regions = measure.regionprops(labelled_max)
    peak_coords = np.zeros(shape=(len(regions), 2), dtype=np.int)
    
    distance_from_mask = distance_transform_cdt(image_copy * (1-mask))

    for i, props in enumerate(regions):
        y0, x0 = props.centroid
        peak = np.array([y0, x0])

        # ensure that there are no peaks off the map (due to rounding)
        peak[peak < 0] = 0
        for j in range(image_copy.ndim):
            if peak[j] > image_copy.shape[j]:
                peak[j] = image_copy.shape[j] - 1
        
        peak_index = tuple(np.round(peak, 0).astype(int)) # indexing with a floating point array sucks, so convert to a more convenient form
        if distance_from_mask[peak_index] > 2*connectivity:
            peak_coords[i, :] = peak
    return peak_coords


def _peak_search_sep_wrapper(firing_map, **kwargs):
    ''' Wrapper around the 'sep' Peak Search method
    Because the 'sep' package is a nightmare to install, and not used for most
    analysis routines, this wrapper allows most users to ignore it
    
    If the user tries to invoke the 'sep' routines, this will try to do so
    If it fails due to ModuleNotFound, it will use the default algorithm instead
    with a warning to the user
    '''
    try:
        import sep
        return _peak_search_sep(firing_map, **kwargs)
    except ModuleNotFoundError:
        raise ModuleNotFoundError("The package 'sep' is missing from your system."\
                " You can invoke an alternative algorithm that does not depend on"\
                " 'sep' by assigning a different value to 'search_method'."\
                " Alternatively, install 'sep' on your system:"\
                " 'pip install sep'")
    
        
        

def _peak_search_sep(firing_map, **kwargs):
    '''Peak search using sep, a Python wrapper for a standard astronomy library.
    sep is typically used to identify astonomical objects in telescope images
    Sep requires copies of the arrays that are C-ordered (Python default is 
    Fortran-ordered, the difference is row vs column-major).
    
    TODO: The behaviour of sep with masks needs to be investigated - where the 
    edge of the image is masked, objects are still sometimes found. Example: 
        {'path': 'N:\\davidcr\\84932\\19032019',
         'basename': '19032019s1',
         'tetrode': 6,
         'cell': 23}
    '''
    import sep
    
    mask = kwargs.get("mask", np.zeros(firing_map.shape, dtype=bool))
    null_background = kwargs.get("null_background", True)
    threshold = kwargs.get("threshold", 0.2)
    
    tmp_firing_map = firing_map.copy('C')
    tmp_mask = mask.copy(order='C')
    
    if null_background:
        bkg = sep.Background(np.zeros_like(tmp_firing_map))
    else:
        bkg = sep.Background(tmp_firing_map, mask=tmp_mask, fw=2, fh=2, \
                     bw=int(tmp_firing_map.shape[0]), bh=int(tmp_firing_map.shape[1]))

    init_fields = sep.extract(tmp_firing_map-bkg, mask=tmp_mask, thresh=threshold, \
                          err=bkg.globalrms)

    peak_coords = np.zeros(shape=(len(init_fields), 2), dtype=np.int)

    for i, props in enumerate(init_fields):
        peak = np.array([props['y'], props['x']])
        peak = np.round(peak)
        # ensure that there are no peaks off the map (due to rounding)
        peak[peak < 0] = 0
        for j in range(firing_map.ndim):
            if peak[j] > firing_map.shape[j]:
                peak[j] = firing_map.shape[j] - 1

        peak_coords[i, :] = peak
    return peak_coords


""" Tests for smoothing"""
import numpy as np
import pytest

from opexebo.general import smooth as func

print("=== tests_general_smooth ===")


###############################################################################
################                MAIN TESTS
###############################################################################

# Want to test both sparse and well occupied maps
# Should also test square vs circle vs linear


def test_invalid_inputs():
    # >2D dimensional data
    with pytest.raises(NotImplementedError):
        d = np.ones((5, 5, 5))
        func(d, sigma=2)
    print("test_invalid_inputs passed")
    return True



"""
Calculate 2D spatial autocorrelation

Calculates 2D autocorrelation (autocorrelogram) of a firing map.

"""

import numpy as np
import opexebo


def autocorrelation(firing_map):
    """Calculate 2D spatial autocorrelation of a firing map.

    Parameters
    ----------
    firing_map: np.ndarray
        NxM matrix, smoothed firing map. map is not necessary a numpy array. 
        May contain NaNs.

    Returns
    -------
    acorr: np.ndarray
        Resulting correlation matrix, which is a 2D numpy array.

    See Also
    --------
    opexebo.general.normxcorr2_general
        
    Notes
    -----
    BNT.+analyses.autocorrelation

    Copyright (C) 2018 by Vadim Frolov
    """

    # overlap_amount is a parameter that is intentionally not exposed to
    # the outside world. This is because too many users depend on it and we
    # do not what everyone to use their own overlap value.
    # Should be a value in range [0, 1]
    overlap_amount = 0.8
    slices = []

    if type(firing_map) != np.ndarray:
        firing_map = np.array(firing_map)

    if firing_map.size == 0:
        return firing_map

    # make sure there are no NaNs in the firing_map
    firing_map = np.nan_to_num(firing_map)

    # get full autocorrelgramn
    aCorr = opexebo.general.normxcorr2_general(firing_map)

    # we are only interested in a portion of the autocorrelogram. Since the values
    # on edges are too noise (due to the fact that very small amount of elements
    # are correlated).
    for i in range(firing_map.ndim):
        new_size = np.round(firing_map.shape[i] + firing_map.shape[i] * overlap_amount)
        if new_size % 2 == 0:
            new_size = new_size - 1
        offset = aCorr.shape[i] - new_size
        offset = np.round(offset/2 + 1)
        d0 = int(offset-1)
        d1 = int(aCorr.shape[i] - offset + 1)
        slices.append(slice(d0, d1))

    return aCorr[tuple(slices)]



"""Tests for accumulate_spatial"""
import numpy as np
import pytest

import opexebo
from opexebo.general import accumulate_spatial as func

print("=== tests_general_accumulate_spatial ===")


def test_invalid_inputs():
    # No `arena_size` keyword
    with pytest.raises(TypeError):
        pos = np.random.rand(100)
        func(pos)
    # Misdefined bins
    with pytest.raises(KeyError):
        pos = np.random.rand(100)
        func(pos, arena_size=1, bin_number=10, bin_width=2.5)
    # Misdefined `limit` keyword
    with pytest.raises(ValueError):
        post = np.random.rand(100)
        func(post, arena_size=1, limits="abc")


def test_1d_input():
    arena_size = 80
    pos = np.random.rand(1000) * arena_size
    bin_width = 2.32
    limits = (np.nanmin(pos), np.nanmax(pos) * 1.0001)
    hist, edges = func(pos, arena_size=arena_size, limits=limits, bin_width=bin_width)
    assert hist.ndim == 1
    assert hist.size == opexebo.general.bin_width_to_bin_number(arena_size, bin_width)
    assert edges.size == hist.size + 1
    assert pos.size == np.sum(hist)

def test_1d2d_input():
    """
    It is possible to give 1d data as a 2d array, i.e. with shape (1, n)
    """
    arena_size = 80
    pos = np.random.rand(1000) * arena_size
    pos = np.expand_dims(pos, 0)
    bin_width = 2.32
    limits = (np.nanmin(pos), np.nanmax(pos) * 1.0001)
    hist, edges = func(pos, arena_size=arena_size, limits=limits, bin_width=bin_width)
    assert hist.ndim == 1
    assert hist.size == opexebo.general.bin_width_to_bin_number(arena_size, bin_width)
    assert edges.size == hist.size + 1
    assert pos.size == np.sum(hist)


def test_2d_input():
    arena_size = np.array((80, 120))
    pos = (np.random.rand(1000, 2) * arena_size).transpose()

    limits = (0, 80.001, 0, 120.001)
    bin_width = 4.3
    hist, (edge_x, edge_y) = func(
        pos, arena_size=arena_size, limits=limits, bin_width=bin_width
    )
    assert edge_x[0] == limits[0]
    assert hist.ndim == 2
    for i in range(hist.ndim):
        # Note: the array is transposed, so the shape swaps order
        #        print(hist.shape)
        #        print(opexebo.general.bin_width_to_bin_number(arena_size, bin_width))
        #        print(edge_x[0], edge_x[1])
        #        print(np.min(pos[0]), np.max(pos[0]))
        assert (
            hist.shape[i]
            == opexebo.general.bin_width_to_bin_number(arena_size, bin_width)[i - 1]
        )
    assert pos.shape[1] == np.sum(hist)


def test_2d_bin_number():
    arena_size = np.array((80, 120))
    pos = (np.random.rand(1000, 2) * arena_size).transpose()
    limits = (0, 80.001, 0, 120.001)
    bin_number = (8, 12)
    hist, (edge_x, edge_y) = func(
        pos, arena_size=arena_size, limits=limits, bin_number=bin_number
    )
    assert edge_x.size == bin_number[0] + 1
    assert edge_y.size == bin_number[1] + 1
    assert pos.shape[1] == np.sum(hist)

    bin_number = 8
    hist, (edge_x, edge_y) = func(
        pos, arena_size=arena_size, limits=limits, bin_number=bin_number
    )
    assert edge_x.size == edge_y.size == bin_number + 1
    assert pos.shape[1] == np.sum(hist)


def test_2d_bin_edges():
    arena_size = np.array((80, 120))
    pos = (np.random.rand(1000, 2) * arena_size).transpose()
    limits = (0, 80.001, 0, 120.001)
    bin_edges = [np.arange(arena_size[i] + 1) for i in range(2)]
    hist, edges = func(pos, arena_size=arena_size, limits=limits, bin_edges=bin_edges)
    for i in range(2):
        assert np.array_equal(edges[i], bin_edges[i])
    # Also test that passing in an array instead of a list works:
    bin_edges = np.array(bin_edges)
    hist, edges = func(pos, arena_size=arena_size, limits=limits, bin_edges=bin_edges)
    for i in range(2):
        assert np.array_equal(edges[i], bin_edges[i])


if __name__ == "__main__":
    test_invalid_inputs()
    test_2d_input()
    test_2d_bin_number()
    test_2d_bin_edges()



"""Provide function for shuffling"""

import numpy as np
import time

def shuffle(times, offset_lim, iterations, **kwargs):
    '''
    Increment the provided time series by a random period of time in order to 
    destroy the correlation between spike times and animal behaviour.
    
    The increment behaves circularly
    
    * Initially, we have two time series, Tracking and SpikeTimes, both with 
      values in the range `[t0, t1]`. 
    * After incrementing the SpikeTimes by T, we have Tracking in `[t0, t1]`
      and SpikeTimes in `[t0+T, t1+T]`.
    * Take all spike times in the range `[t1, t1+T]` and map back to `[t0, t0+T]`
      by subtracting `(t1-t0)`
    
    The end result should be a series of times also in the range [t0, t1], 
    with the same intervals between times, but the exact value of those times
    has changed. if tracking_range not provided, then two spikes (previously 
    first, last) will noe have zero spacing
    
    The random numbers are drawn from a pseudorandom, uniform, distribution in 
    the range `[offset_lim, t1-offset_lim]`
        
    
    Parameters
    ----------
    times: np.ndarray
        1D array of times to be shuffled
    offset_lim: float
        Defines the range of values by which each iteration can be offset. Each
        iteration will have an offset in `[offset_lim, max(times)-offset_lim]`
    iterations: int
        How many copies of times should be returned (each copy incremented by a
        random offset limited by `offset_lim`)
    tracking_range: array_like, optional
        The time range over which tracking behaviour exists, defining the
        times at which spike indexes are looped back on themselves
        This can be provided either as a 2-element tuple `[t0, t1]`, or as 
        the entire list of tracking timestamps. In each case, the min and 
        max values are used as `t0`, `t1`.
        If no values are provided, then the first and last spike times
        are used. This is not desirable behaviour, since it will then guarantee 
        that in the shuffled output, two spikes will occur simultaneously
    debug: bool, optional
        Enable additional debugging output
    
    Returns
    -------
    output: np.ndarray
        iterations x N array of times. To access a single iteration, output[i,:] or output[i]
    increments : np.ndarray
        1D array of offset values used.
    
    Notes
    --------
    BNT.+scripts.shuffling around line 350
    
    Copyright (C) 2019 by Simon Ball
    '''
    
    # Check values
    if type(times) != np.ndarray:
        times = np.array(times)
    if times.ndim != 1:
        raise ValueError("You must provide a 1D array of times. You provided a"\
                         " %d-dimensional array" % times.ndim)
    if times.size == 0:
        raise ValueError("Your spike_times array is empty")
    if np.isnan(times).any():
        raise ValueError("You have NaN values in your times array")
    debug = kwargs.get('debug', False)
    tr = kwargs.get('tracking_range', None)
    try:
        if tr is not None:
            t0 = np.nanmin(tr)
            t1 = np.nanmax(tr)
            if (t0 > np.nanmin(times)) or (t1 < np.nanmax(times)):
                raise ValueError("Your times cover a larger span of time than your"\
                                 " tracking information")
    except ValueError as e:
        print(e)
        print(times)
    else:
        t0 = np.nanmin(times)
        t1 = np.nanmax(times)
        
    
    if offset_lim >= 0.5*(t1-t0):
        raise ValueError("offset_lim must be less than half of the time-span"\
                         " covered. You provided %.2g and a time span of %.2g"\
                         % (offset_lim, t1-t0))

    
    # Initialise Numpy's random number generator with a new seed based on the 
    # user's local computer OS methods
    t_min = t0 + offset_lim
    t_max = t1 - offset_lim
    
    
    increments = np.random.RandomState().rand(iterations) # Uniformmly distrbuted in [0, 1]
    increments = t_min + (increments * (t_max-t_min)) # uniformly distributed in [t_min, t_max]
    
    num_spikes = np.size(times)
    
    if debug:
        print("Number of spikes provided: %d" % num_spikes)
        print("Spikes in time range [%d, %d]" % (t0, t1))
        print("Iterations requested: %d" % iterations)
        print("Increments in range [%.2f, %.2f]" % (np.min(increments), np.max(increments)))
    
    
    # Generate 2D arrays by repeating over the correct axis
    # Want to duplicate the whole list of times for each repeat
    # Want to duplicate the *same* increment for each repeat. Hence the transpose. 
    all_times = np.repeat(times[:,np.newaxis], iterations, axis=1)              
    all_inc = np.repeat(increments[:, np.newaxis], times.size, axis=1).transpose()
    new_times = all_times + all_inc
    
    
    
    to_shift = (new_times > t1) # Binary array of the elements that need to be moved to the beginning. Each column is different
    time_idx = num_spikes - np.count_nonzero(to_shift, axis=0) # indicies of the first time in each iteration which must be moved
    if debug:
        print("New time range: [%d, %d]" % (np.min(new_times), np.max(new_times)))
        print("Indexes exceeding value t1: %s" % to_shift)
        
    output = np.zeros(new_times.shape)

    a0 = time.time()
    for i in range(iterations):
        # Circular buffer. We have alread identified the index at which elements 
        # need to be moved (time_idx[i]). So take those end elements, and move 
        # them to the start
        # Turns out, this is actually faster than fancy array indexing with np.ix_(time_idx, iter_idx), by nearly 100x (!) in testing
        
        # Elements that are still within the time range[t0+T, t1]
        a = new_times[0:time_idx[i], i]
        
        # Elements that are in the time range [t1, t1+T]
        # len(b) == num_spikes-time_idx[i]
        b = new_times[time_idx[i]:, i]
        
        # Elements in b should have their values reduced to fit in the range [t0, t0+T]        
        b = b - t1 + t0
        
        # Insert elements into output in the correct order: b first, then a filling in the blanks after
        output[0:num_spikes-time_idx[i], i] = b
        output[num_spikes-time_idx[i]:, i] = a

    a1 = time.time()
    if debug:
        print("Final time range: [%d, %d]" % (np.min(output), np.max(output)))
        print("method 1 took %dms" % ((a1-a0)*1000))
        print(np.diff(times))
        print(np.diff(output[:,0]))
    
    # Want to return such that each row of the output is a single time-shifted iteration
    # This will allow a single iteration of ShuffledUnitSikeTimes to be accessed as output[i]
    output = output.transpose()
    
    return output, increments
    
#if __name__ == '__main__':
#    times = [2, 12, 27, 54, 82, 113, 115, 207, 300]
#    t_min = 20
#    iterations = 3
#    s = shuffle(times, t_min, iterations, debug=False, tracking_range=(0,349))
#    #print(s)

    


import numpy as np
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    # Suppress the ConfigurationMissingWarning that Astropy triggers
    from astropy.convolution import convolve


def calc_speed(t, x, y=None, moving_average=None):
    """
    Calculate the speed of an animal given a list of times and positions

    Speed is calculated element-wise, with an optional moving-average filter.
    Speed is not checked for non

    Parameters
    ----------
    t : np.ndarray
        1d array containing time stamps
    x, y : np.ndarray
        1d arrays containing positions. `y` is optional and may be excluded for
        1d datasets
    moving_average: int, optional
        Apply a moving-average filter of this size to the speed data
        To exclude this filter, leave blank or set [None] (default)
    
    Returns
    -------
    np.ndarray
        1d array containing speed in units [position_unit / time_unit]
    """
    if y is None:
        y = np.zeros(t.size)
    arg_names = ["t", "x", "y"]
    for i, arg in enumerate((t, x, y)):
        if not isinstance(arg, (list, tuple, np.ndarray, np.ma.MaskedArray)):
            raise ValueError(
                "{} must be an iterable, not  {}".format(arg_names[i], type(arg))
            )
        if not isinstance(arg, (np.ndarray, np.ma.MaskedArray)):
            arg = np.array(arg)
        if not arg.ndim == 1:
            raise ValueError(
                "{} must be 1-dimensional, not {}-dimensional".format(
                    arg_names[i], arg.ndim
                )
            )
        if not arg.size == t.size:
            raise ValueError(
                "{} must be the same length as `t` ({}, {})".format(
                    arg_names[i], arg.size, t.size
                )
            )

    if moving_average is not None:
        if not isinstance(moving_average, int):
            raise ValueError(
                "`moving_average` must be None or type `int`, not {}".format(
                    moving_average
                )
            )
        if moving_average < 3 or moving_average % 2 != 1:
            raise ValueError(
                "`moving_average` must be a positive odd integer greater than 1"
            )

    d_x = np.diff(x)
    d_y = np.diff(y)
    d_t = np.diff(t)
    speed = abs(np.sqrt(np.square(d_x) + np.square(d_y)) / d_t)
    new_speed = np.zeros(speed.size + 1)
    new_speed[1:] = speed
    new_speed[0] = speed[0]

    if moving_average is not None:
        kernel = np.ones(moving_average) / moving_average
        new_speed = convolve(new_speed, kernel, boundary="extend")
    return new_speed



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 18 11:14:37 2019

@author: simoba
"""

import numpy as np

def theta_modulation_index(spike_times, **kwargs):
    '''
    Calculate the 1-dimensional autocorrelation of the spike train and use it
    as a proxy score for how how reliably a cell fires at the same phase of
    the theta frequency.
    
    The measure is calculated as described in publication (with small modifications)
    Cacucci et al, Theta-Modulated Place-by-Direction Cellsin the Hippocampal
    Formation in the Rat, Journal of Neuroscience, 2004.
    doi: 10.1523/jneurosci.2635-04.2004
    
    The calculation is very similar in practice, although somewhat less 
    sophisicated, than the standard g(2) autocorrelation calculation.
    
    For each spike, consider all spikes that follow within some of length P
        Calculate the time delta to all spikes within P and bin as a histogram
    Convert from number of occurrences to rate of occurrences (optional)
    Calculate Index as the contrast of the autocorrelation at Tau=(50,70ms) to
    Tau=(100,140ms) (assumed theta trough and peak respectively)
    
    This indicator fails in two key areas:
        * It indicates nothing about the phase of theta, i.e. this is a scalar rather than vector calculation
        * It _assumes_ the value of theta for the animal in question, rather than calculating it directly
    
    Parameters
    ----------
    spike_times : np.ndarray
        Array of times at which spikes for a cell are detected, in [seconds]
    
    Returns
    -------
    theta_modulation_index : float
        Contrast of autocorrelation
    hist : np.ndarray
        Histogram of time_delta to subsequent spikes within the next 0.5s
    bins : np.ndarray
        Bin edges of the histogram in [s]
    
    '''
    
    if spike_times.ndim != 1:
        raise ValueError("spike_times must be a 1D array")
    
    
    # We start with 1D array spike_times
    # The neatest solution would be to calculate a 2D array of time_deltas, and
    # then take a 1D histogram of that cutoff at tau=0.
    # However, that will rapidly run into memory constraints, so we probably
    # need to iterate over each spike in turn
    
    lower = 0
    upper = 0.5
    bin_size = 5e-3
    bins = np.linspace(lower, upper, int(1/bin_size))
    hist = np.zeros(bins.size-1)
    
    for i, t in enumerate(spike_times):
        # Identify the section of spike_times that is relevant to us here, i.e.  i:j
        try:
            j = i+np.where(spike_times[i:] > t+upper)[0][0]
            time_delta = spike_times[i+1:j] - t
            
            hist_i, _ = np.histogram(time_delta, bins=bins, range=(lower, upper))
            hist += hist_i
        except IndexError:
            # When we are within <upper> seconds of the end of the recording,
            # the above will fail (because j will be meaningless)
            # In any case, we lose information in the final, tiny section
            # So discard it completely
            break
        
    # hist is now the sum of all time_deltas (i.e. tau) in the range (lower, upper)
    # theta_modulation_index is defined as the contrast between two specific, 
    # arbitrary, time windows
    # Arbitrary values taken from the paper cited above. 
    
    arbitrary_trough = np.where((50e-3 <= bins) & (bins < 70e-3))
    arbitrary_peak = np.where((100e-3 <= bins) & (bins < 140e-3))
    
    # Only provide a value if there are at least _some_ spikes within each window
    min_spikes_for_contrast = 5
    peak = np.sum(hist[arbitrary_peak])
    trough = np.sum(hist[arbitrary_trough])
    if peak >= min_spikes_for_contrast and trough >= min_spikes_for_contrast:
        contrast = (peak-trough) / (peak+trough)
    else:
        contrast = np.nan
    
    return contrast, hist, bins


""" Tests for autocorrelogram"""
import inspect
import numpy as np

from opexebo.analysis import angular_occupancy as func

print("=== tests_analysis_angular_occupancy ===")


###############################################################################
################                MAIN TESTS
###############################################################################


def test_angular_random_data():
    """Doesn't test that meaningful results are produced, only that something is"""
    n = 1000
    time = np.arange(n)
    angles = np.random.rand(n)
    a, b, c = func(time, angles)
    print(f"{inspect.stack()[0][3]} passed")
    return True


def test_angular_zero_data():
    """Doesn't test that meaningful results are produced, only that something is"""
    n = 1000
    time = np.arange(n)
    angles = np.zeros(n)
    a, b, c = func(time, angles)
    print(f"{inspect.stack()[0][3]} passed")
    return True


# if __name__ == '__main__':
#    test_angular_random_data()
#    test_angular_zero_data()



"""
Provide function for gridness score calculation.
"""

import numpy as np

from scipy.spatial.distance import cdist
from scipy.stats import pearsonr
from skimage import transform
import opexebo
import opexebo.defaults as default
import opexebo.errors as err


# This is included as the return result of an invalid autocorrelogram
# It replaces an earlier mixture of output types that performed calcaultions 
# on a null-acorr instead
INVALID_OUTPUT = (np.nan, {'grid_spacings': np.array([np.nan, np.nan, np.nan]),
  'grid_spacing': np.nan,
  'grid_orientations': np.array([np.nan, np.nan, np.nan]),
  'grid_orientations_std': np.nan,
  'grid_orientation': np.nan,
  'grid_positions': np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]),
  'grid_ellipse': np.nan,
  'grid_ellipse_aspect_ratio': np.nan,
  'grid_ellipse_theta': np.nan})


def grid_score(aCorr, **kwargs):
    """Calculate gridness score for an autocorrelogram.

    Calculates a gridness score by expanding a circle around the centre field
    and calculating a correlation value of that circle with it's rotated versions.
    The expansion is done up until the smallest side of the autocorrelogram.
    The function may also calculate grid statistics.

    Gridness score value by itself is calculated as a maximum over a sliding mean
    of expanded circle. The width of the sliding window is given by a variable
    numGridnessRadii. This is done in order to keep gridness score the same as
    historical values (i.e. older versions of gridness score).

    Parameters
    ----------
    acorr: np.ndarray
        A 2D autocorrelogram.
    
    Other Parameters
    ----------------
    min_orientation: int
        See function "grid_score_stats"
    search_method: str
        Peak searching method, currently limited to either `default` or `sep`
    bin_width: float
        Size of the bins. Distance units will be returned in the same units
        If not provided, distance units will be retuned in units [bins]

    Returns
    -------
    grid_score: float
        Always returns a gridness score value. It ranges from -2 to 2. 2 is
        more of a theoretical bound for a perfect grid. More practical value for
        a good grid is around 1.3. If function can not calculate a gridness
        score, NaN value is returned.
    grid_stats: dictionary
        grid_spacings: np.array
            Spacing of three adjacent fields closest to center in autocorr
            (in [bins] if keyword `bin_width` is not given)
        grid_spacing: float
            Nanmean of 'spacings' (in [bins] if keyword `bin_width` is not
            given)
        grid_orientations: np.array
            Orientation of three adjacent fields closest to center in autocorr
            (in [degrees])
        grid_orientations_std: float
            Standard deviation of orientations % 60
        grid_orientation: float
            Orientation of grid in [degrees] (mean of fields of 3 main axes)
        grid_positions: np.array
            [y,x] coordinates of six fields closest to center
        grid_ellipse: np.array
            Ellipse fit returning 
            [x coordinate, y coordinate, major radius, minor radius, theta]
        grid_ellipse_aspect_ratio: float
            Ellipse aspect ratio (major radius / minor radius)
        grid_ellipse_thet: float
            Ellipse theta (corrected according to previous BNT standard) in [degrees]

    See Also
    --------
    opexebo.analysis.placefield

    Notes
    -----
    BNT.+analyses.gridnessScore

    Copyright (C) 2018 by Vadim Frolov, (C) 2019 by Simon Ball, Horst Obenhaus
    """
    debug = kwargs.get("debug", False)

    # normalize aCorr in order to find contours
    aCorr = aCorr / aCorr.max()
    centre = -0.5 + np.array(aCorr.shape)/2 # centre : also [y, x]
    cFieldRadius = int(np.floor(_findCentreRadius(aCorr, **kwargs)))

    if cFieldRadius == 0:
        if debug:
            print("Terminating due to invalid cFieldRadius")
        return INVALID_OUTPUT

    halfHeight = np.ceil(aCorr.shape[0]/2)
    halfWidth  = np.ceil(aCorr.shape[1]/2)
    heightIndices = np.arange(aCorr.shape[0])
    widthIndices  = np.arange(aCorr.shape[1])

    # Define radii that will be iterated over for the gridness score
    # outer bound is defined by the minimum of autocorrelogram's dimensions
    # this is need for rectangular autocorrelograms.
    outerBound = int(np.floor(np.min(np.array(aCorr.shape)/2)))
    if outerBound < cFieldRadius:
        if debug:
            print("Terminating due to invalid outerBound"\
                  f" ({outerBound} < {cFieldRadius})")
        return INVALID_OUTPUT
        
    radii = np.linspace(max(3, cFieldRadius+1), outerBound, outerBound-cFieldRadius)
    radii = radii.astype(int)
    numSteps = len(radii)
    if numSteps < 1:
        if debug:
            print("Terminating due to invalud numSteps")
        return INVALID_OUTPUT

    rotAngles_deg = np.arange(30, 151, 30)  # 30, 60, 90, 120, 150
    rotatedACorr = np.zeros(
            shape=(aCorr.shape[0], aCorr.shape[1], len(rotAngles_deg)),
            dtype=float)
    # we get rotated maps here as it is a heavy operation
    for n, angle in enumerate(rotAngles_deg):
        rotatedACorr[:, :, n] = transform.rotate(aCorr, angle,
                                            preserve_range=True, clip=False)
    rr, cc = np.meshgrid(widthIndices, heightIndices, sparse=False)
    # This is needed for compatibility with Matlab's code
    rr += 1
    cc += 1
    mainCircle = np.sqrt(np.power((rr - halfWidth), 2) + np.power(cc-halfHeight, 2))
    innerCircle = mainCircle > cFieldRadius


    GNS = np.zeros(shape=(numSteps, 2), dtype=float)
    for i, radius in enumerate(radii):
        mask = innerCircle & (mainCircle < radius)
        aCorrValues = aCorr[mask]

        rotCorr = np.zeros_like(rotAngles_deg).astype(float)
        for j, angle in enumerate(rotAngles_deg):
            rotatedValues = rotatedACorr[mask, j]
            r, _ = pearsonr(aCorrValues, rotatedValues)
            rotCorr[j] = r
        GNS[i, 0] = np.min(rotCorr[[1, 3]]) - np.max(rotCorr[[0, 2, 4]])
        GNS[i, 1] = radius

    # find the greatest gridness score value and radius
    gscoreInd = np.argmax(GNS[:,0])

    numGridnessRadii = 3
    numStep = max(numSteps - numGridnessRadii, 1) # minimum value 1

    if numStep == 1:
        gscore = np.mean(GNS, axis=0)[0]
    else:
        meanGridness = np.zeros(numStep)
        for ii in range(numStep):
            meanGridness[ii] = np.nanmean(GNS[ii:ii+numGridnessRadii, 0])
        gscore = np.max(meanGridness)

    '''Then calculate stats about the autocorrelogram'''
    # Mask center field and fringes of autocorrelogram > best grid score radius
    mask_outwards = _circular_mask(aCorr, radii[gscoreInd]*1.25, 'outwards', centre)
    mask_center   = _circular_mask(aCorr, cFieldRadius*1.5, 'inwards', centre)
    mask = mask_outwards + mask_center

    grid_stats = grid_score_stats(aCorr, mask, centre, **kwargs)

    return gscore, grid_stats

#########################################################
################        Helper Functions
#########################################################


def grid_score_stats(aCorr, mask, centre, **kwargs):
    '''
    Calculate spatial characteristics of grid based on 2D autocorr

    Parameters
    ----------
    aCorr : np.array
        2D Autocorrelation
    mask : np.array
        Mask (masked=True) of shape aCorr for masking center field and
        fringes of aCorr above best grid score radius
    centre : np.array
        Centre coordinate [y,x]
    **kwargs :
        min_orientation : int
            Minimum difference in degrees that two neighbouring fields
            detected in 2D autocorrelation must have. If difference is
            below this threshold, discard the field that has larger
            distance from center
        search_method : str
            Peak searching method, currently limited to either `default` or `sep`
        bin_width : float
            Size of the bins. Distance units will be returned in the same units
            If not provided, distance units will be retuned in units [bins]

    Returns
    -------
    grid_stats : dictionary
        grid_spacings              : np.array
            Spacing of three adjacent fields closest to center in autocorr
            (in [bins])
        grid_spacing                : float
            Nanmean of 'spacings' in [bins]
        grid_orientations           : np.array
            Orientation of three adjacent fields closest to center in autocorr
            (in [degrees])
        grid_orientations_std       : float
            Standard deviation of orientations % 60
        grid_orientation            : float
            Orientation of grid in [degrees] (mean of fields of 3 main axes)
        grid_positions              : np.array
            [y,x] coordinates of six fields closest to center
        grid_ellipse                : np.array
            Ellipse fit returning 
            [x coordinate, y coordinate, major radius, minor radius, theta]
        grid_ellipse_aspect_ratio   : float
            Ellipse aspect ratio (major radius / minor radius)
        grid_ellipse_theta          : float
            Ellipse theta (corrected according to previous BNT standard) in [degrees]
    '''

    # Get kwargs
    debug = kwargs.get('debug', False)
    bin_width = kwargs.get("bin_width", 1) # if not provided, use a 1:1 match from bins to return units
    min_orientation = kwargs.get('min_orientation', default.min_orientation)
    search_method = kwargs.get("search_method", default.search_method)
    min_orientation = np.radians(min_orientation)
    if debug:
        print(f"Search method: {search_method}")
        print('Min orientation: {} degrees'.format(np.degrees(min_orientation)))
    

    # Initialise default output in case stats are uncalculable
    gs_ellipse_theta    = np.nan
    gs_ellipse          = np.nan
    gs_aspect_ratio     = np.nan
    gs_orientations_std = np.nan
    gs_orientation      = np.nan
    gs_positions        = np.full(6, fill_value = np.nan, dtype=float)
    gs_orientations     = np.full(3, fill_value = np.nan, dtype=float)
    gs_spacings         = np.full(3, fill_value = np.nan, dtype=float)

    # Find fields in autocorrelogram
    all_coords = opexebo.general.peak_search(aCorr, mask=mask, search_method=search_method,
                                             null_background=True, threshold=0.1, get_maxima=True)
    if debug:
        import matplotlib.pyplot as plt
        plt.figure()
        plt.imshow(aCorr)
        plt.scatter(centre[1], centre[0], s=600, marker="x", color="black")
        for field_no, coord in enumerate(all_coords):
            plt.scatter(coord[1], coord[0], s=300, marker='x', color='red')
            plt.text(coord[1]+3, coord[0], field_no, label='Center')
        plt.title("All local maxima in acorr")
        
        

    if all_coords.shape[0] >= 6:
        # Calculate orientation and distance of all local maxima to center
        # np.arctan2 gives angles in radians relative to the horizontal axis in the range [-pi, pi]
        # A vector along the horizontal axis from (0,0) to (0,-1) has angle pi.
        # np.arctan2 - accepts arguments (y, x), and gives the angles in each case from (0,0) to (xi, yi)
        orientation = np.arctan2(all_coords[:,0] - centre[0], all_coords[:,1] - centre[1]) # in radians
        distance = np.sqrt(np.square(all_coords[:,0]-centre[0]) + np.square(all_coords[:,1]-centre[1]))

        # Where two fields have a very similar orientation, discard the more distant one
        orient_distsq = np.abs(_circ_dist2(orientation))
        close_fields = orient_distsq < min_orientation
        close_fields = np.triu(close_fields, 1) # Upper triangle only - set lower triangle to zero
                                                # k=1: +1 offset from diagonal: set diagonal to zero too
        to_del = []
        for row,col in np.argwhere(close_fields):
            if distance[row] > distance[col]:
                to_del.append(row)
            else:
                to_del.append(col)

        distance    = np.delete(distance, to_del)
        orientation = np.delete(orientation, to_del)
        all_coords  = np.delete(all_coords, to_del, axis=0)

        # First sort by distance and take first 6 fields
        sorted_ids_dist = np.argsort(distance)[:6] # 6 closest fields
        positions = all_coords[sorted_ids_dist]
        distance = distance[sorted_ids_dist]
        orientation = orientation[sorted_ids_dist]
        
        # ... then re-sort remaining fields by angle
        sorted_ids_ang = np.argsort(orientation)

        ################# GATHER OUTPUT #################
        gs_positions = positions[sorted_ids_ang]
        # For grid orientation and spacing take only 3 out of 6 neighbouring fields
        # Convert from distance in bins to distance in units
        gs_spacings = distance[sorted_ids_ang][:3] * bin_width
        gs_orientations = np.degrees(orientation[sorted_ids_ang][:3]) % 180
        # Work out mean orientation of grid. Take standard deviation as quality marker
        gs_orientation, gs_orientations_std = _extract_grid_orientation(gs_orientations)


        if debug:
            import matplotlib.pyplot as plt
            aCorr_masked = np.ma.masked_where(mask, aCorr.copy())
            plt.figure()
            plt.imshow(aCorr_masked)
            plt.scatter(centre[1],centre[0], s=600, marker='x', color='black')
            for field_no, coord in enumerate(gs_positions):
                plt.scatter(coord[1], coord[0], s=300, marker='x', color='red')
                plt.text(coord[1]+3, coord[0], field_no, label='Center')
            plt.title('Masked autocorr + 6 remaining fields')

        # Fit an ellipse to those remaining fields:
        if len(gs_positions) > 2:
            try:
                gs_ellipse =  opexebo.general.fit_ellipse(gs_positions[:,1], gs_positions[:,0])
                gs_ellipse_theta = np.degrees(gs_ellipse[4]+np.pi)%360
                # The +pi term was included in the original BNT, I have kept it to
                # maintain consistency with past results.
                gs_aspect_ratio = gs_ellipse[2]/gs_ellipse[3] # Major radius / Minor radius
            except np.linalg.LinAlgError as e:
                print(f"LinAlgError: {e}")
                print(f"Retuning NaN ellipse stats")

    else:
        if debug: 
            print('Not enough fields detected ({})'.format(len(all_coords)))
            

    grid_stats = {'grid_spacings': gs_spacings,
                  'grid_spacing': np.nanmean(gs_spacings),
                  'grid_orientations': gs_orientations,
                  'grid_orientations_std': gs_orientations_std,
                  'grid_orientation': gs_orientation,
                  'grid_positions': gs_positions,
                  'grid_ellipse': gs_ellipse,
                  'grid_ellipse_aspect_ratio': gs_aspect_ratio,
                  'grid_ellipse_theta': gs_ellipse_theta}
    return grid_stats


def _circular_mask(image, radius, polarity='outwards', center=None):
        '''
        Given height and width, create circular mask around point with defined radius
        Polarity:
            'inwards' : True inside,  False outside
            'outwards': True outside, False inside
        '''
        h = image.shape[0]
        w = image.shape[1]

        if center is None:
            center = [int(h/2), int(w/2)]

        Y, X = np.ogrid[:h, :w]
        dist_from_center = np.sqrt(np.power(X - center[1],2) + np.power(Y-center[0],2))
        if polarity.lower() == 'inwards':
            mask = dist_from_center <= radius
        elif polarity.lower() == 'outwards':
            mask = dist_from_center >= radius
        else:
            raise err.ArgumentError('Polarity "{}" not defined'.format(polarity))
        return mask

def _draw_ellipse(x, y, rl, rs, theta):
    import matplotlib.pyplot as plt
    from matplotlib.patches import Ellipse
    from matplotlib import transforms
    theta = (theta%(2*np.pi))# + np.pi
    ell = Ellipse((0,0), width=rs*2, height=rl*2, facecolor=(1,0,0,0.2),
                  edgecolor=(1,1,1,0.75))
    ax = plt.gca()
    transf = transforms.Affine2D().rotate(theta+np.pi/2).translate(x, y)
    ell.set_transform(transf + ax.transData)
    ax.add_patch(ell)


def _plotContours(img, contours):
    import matplotlib.pyplot as plt
    _, ax = plt.subplots()
    ax.imshow(img, cmap='jet', origin='lower')

    centroids = np.zeros(shape=(len(contours), 2), dtype=float)
    radii = np.zeros(shape=(len(contours), 1), dtype=float)
    for n, contour in enumerate(contours):
        x = contour[:, 1]
        y = contour[:, 0]
        ax.plot(x, y, linewidth=2)
        # mean x and y
        centroid = (sum(x) / len(contour), sum(y) / len(contour))
        radius = [np.mean([np.sqrt(np.square(x-centroid[0]) + np.square(y-centroid[1]))])]

        centroids[n] = centroid
        radii[n] = radius

        ax.text(centroid[0], centroid[1], str(n))
        ax.plot((centroid[0], centroid[0]+radius), (centroid[1], centroid[1]), linewidth=1)

    ax.axis('image')
    ax.set_xticks([])
    ax.set_yticks([])
    plt.show()
    return radii, centroids



def _extract_grid_orientation(orientations):
    '''
    Extract grid orientation based on angular difference
    of autocorrelation (aCorr) field coordinates to 60 degree axes. 

    Parameter
    ---------
    orientations      : np.array
                        (Raw) aCorr field angles in degrees 

    Returns
    -------
    orientation       : float
                        Grid orientation in degrees (average)
    orientation_std   : float
                        Standard deviation over grid field
                        orientations
    '''


    orientations = orientations % 60 
    corr_orientations = []
    for orient in orientations: 
        # For every angle extract min to 60 deg
        diff_60 = orient - 60 
        if np.abs(diff_60) < np.abs(orient): 
            corr_orientations.append(diff_60)
        else:
            corr_orientations.append(orient)

    corr_orientations = np.array(corr_orientations)
    # Check 30 degree flips 
    if np.mean(np.abs(np.abs(corr_orientations) - 30)) < np.mean(np.abs(corr_orientations)):
        # Yes, angles close to 30 degrees (flipping axis)
        # Try to reach consensus
        if np.median(corr_orientations) < 0: # Make everything negative
            corr_orientations = np.negative(corr_orientations, where=corr_orientations>0, out=corr_orientations)
        else: # Make everything positive
            corr_orientations = np.negative(corr_orientations, where=corr_orientations<0, out=corr_orientations)

    # Extract average and standard deviation
    orientation     = np.nanmean(corr_orientations)
    orientation_std = np.nanstd(corr_orientations)

    # Test / correct orientation 
    if np.argmin([np.abs(orientation-60), np.abs(orientation)]) == 0:
        orientation -= 60

    return orientation, orientation_std


def _circ_dist2(X):
    '''Given a 1D array of angles, find the 2D array of pairwise differences
    Based on https://github.com/circstat/circstat-matlab/blob/master/circ_dist2.m'''
    x = np.outer(np.exp(1j*X), np.ones(X.size)) # similar to meshgrid, but simpler to implement
    y = np.transpose(x)
    return np.angle(x/y)



def _polyArea(x, y):
    '''Polygon area, from 
    https://stackoverflow.com/questions/24467972/calculate-area-of-polygon-given-x-y-coordinates
    Seems to be the same as Matlab's polyarea'''
    return 0.5*np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


def _contourArea(contours, i):
    contour = contours[i]
    x = contour[:, 1]
    y = contour[:, 0]
    area = _polyArea(x, y)
    return area


def _findCentreRadius(aCorr, **kwargs):
    debug = kwargs.get("debug", False)
    search_method = kwargs.get("search_method", default.search_method)
    halfHeight = np.ceil(aCorr.shape[0]/2)
    halfWidth = np.ceil(aCorr.shape[1]/2)
    peak_coords = np.ones(shape=(1, 2), dtype=np.int)
    peak_coords[0, 0] = halfHeight-1
    peak_coords[0, 1] = halfWidth-1
    fields = opexebo.analysis.place_field(aCorr, min_bins=5, min_peak=0, min_mean=0, init_thresh=.95, \
                                         peak_coords=peak_coords, search_method=search_method)[0] # Fix all input args for now
    if fields is None or len(fields) == 0:
        if debug:
            print("Terminating _findCentreRadius due to no fields")
        return 0
    elif debug:
        print(f"Fields found: {len(fields)}")
    else:
        pass

    peak_coords = np.ndarray(shape=(len(fields), 2), dtype=np.integer)
    areas = np.ndarray(shape=(len(fields), 1), dtype=np.integer)
    for i, field in enumerate(fields):
        peak_rc = field['peak_coords']
        peak_coords[i, 0] = peak_rc[0]
        peak_coords[i, 1] = peak_rc[1]
        areas[i] = field['area']

    aCorrCentre = np.zeros(shape=(1, 2), dtype=float)
    aCorrCentre[0] = aCorr.shape[0]
    aCorrCentre[0, 1] = aCorr.shape[1]
    aCorrCentre = np.ceil(aCorrCentre/2)

    # index of the closest field to the centre
    closestFieldInd = 0
    if len(peak_coords) >= 2:
        # get all distances and check two minimum of them
        distancesToCentre = cdist(peak_coords, aCorrCentre)
        sortInd = np.squeeze(np.argsort(distancesToCentre, axis=0))

        closestFieldInd = sortInd[0]
        twoMinDistances = distancesToCentre[sortInd[:2]]

        areFieldsClose = np.abs(twoMinDistances[0] - twoMinDistances[1])[0] < 2
        if areFieldsClose:
            # two fields with close middle point. Let's select one with minimum area
            indices_to_test = sortInd[:2]
            min_ind = np.argmin(areas[indices_to_test])
            closestFieldInd = indices_to_test[min_ind]

    radius = np.floor(np.sqrt(areas[closestFieldInd] / np.pi))
    
    if debug:
        print("radius is {}".format(radius))
        
    return radius





""" Tests for turning curve"""
import numpy as np
import inspect

from opexebo.analysis import tuning_curve as func

print("=== tests_analysis_turning_curve ===")


###############################################################################
################                MAIN TESTS
###############################################################################


def test_random_data():
    n = 1000
    bw = 15
    spike_angles = np.random.rand(n) * 2 * np.pi
    angular_occupancy = np.random.rand(int(360 / bw))
    func(angular_occupancy, spike_angles, bin_width=bw)
    print(f"{inspect.stack()[0][3]} passed")
    return True


# if __name__ == '__main__':
#    test_random_data()




import numpy as np

def circular_mask(axes, diameter, **kwargs):
    '''Given a set of axes, construct a circular mask defining pixels as within
    or without a circular arena.
    
    A pixel is assumed to be inside the circle iff the centre-point is within
    the radius of the circle. For small radii, this will result in including
    pixels with <50% of their area inside. 
    
    Parameters
    ----------
        axes: list of np.ndarray
            [`x`, `y`] - pair of nd-arrays defining the edges of the pixels. This
            is the return value from np.histogram2d, for instance. 
            Each one is size+1 relative to the number of pixels
            This function assumes that each dimension has a CONSTANT bin width,
            although each dimension may have its own bin width. 
       diameter: float
           diameter of the circle
       origin: list of floats
           [`x`, `y`] co-ordinates of the centre of the arena. Optional,
           defaults to `(0,0)`
    
    Returns
    -------
    in_field: np.ndarray
        2D boolean array of dimensions `(axes[0].size-1, axs[1].size-1)`
        Values are `true` if INSIDE circle and `false` if OUTSIDE
    distance_map: np.ndarray
        2D float array of same size. Values are the distance of the centre of
        the pixel from the origin
    angular_map: np.ndarray
        2D float array of the same size. Values are the angle from the origin
        to the pixel, in degrees. Angles are given with 0° pointing to increasing
        `y`, clockwise. 
    '''
    origin = kwargs.get("origin", (0,0))
    radius = diameter / 2
    bin_width = [np.mean(np.diff(ax)) for ax in axes]
    x_centres = axes[0][:-1] + (bin_width[0] / 2) - origin[0]
    y_centres = axes[1][:-1] + (bin_width[1] / 2) - origin[1]
    X, Y = np.meshgrid(x_centres, y_centres)
    distance_map = np.sqrt(np.power(X,2) + np.power(Y,2))
    angular_map_rad = np.arctan2(Y, X).transpose()
    angular_map_deg = np.degrees(angular_map_rad) % 360
    in_field = (distance_map <= radius)
    return in_field, distance_map, angular_map_deg


'''
Exception classes for the Opexebo library
'''

# ----- Level 1 -----
class OpexeboError(Exception):
    '''
    Base class for exceptions specific to Opexebo
    '''


# ----- Level 2 -----
class ArgumentError(OpexeboError):
    '''
    Incorrect argument given in some way
    '''

class DimensionMismatchError(OpexeboError):
    """
    Raised if 1D and 2D data is mixed
    """

class SpeedBandwidthError(OpexeboError):
    """
    Raised if the Speedscore bandwidth calculation fails
    """



"""Tests for RateMap"""
import numpy as np
import pytest

import opexebo
from opexebo.analysis import rate_map as func
from opexebo import errors as err

print("=== tests_analysis_rateMap ===")


###############################################################################
################                MAIN TESTS
###############################################################################


def test_rmap_invalid_inputs():
    # Mismatched dimensions
    # 2d time map but 1d spike pos
    with pytest.raises(err.DimensionMismatchError):
        tmap = np.ones((10, 10))
        time = np.arange(100)
        spikes_x = np.random.rand(100)
        spikes_tracking = np.array((time, spikes_x))
        func(tmap, spikes_tracking, arena_size=1)
    # 1d time map but 2d spike_pos
    with pytest.raises(err.DimensionMismatchError):
        tmap = np.ones(10)
        time = np.arange(100)
        spikes_x = np.random.rand(100)
        spikes_y = np.random.rand(100)
        spikes_tracking = np.array((time, spikes_x, spikes_y))
        func(tmap, spikes_tracking, arena_size=(10, 10))
    # invalid input types
    with pytest.raises(err.ArgumentError):
        tmap = (1, 2, 3, 4)  #! tmap should be an ndarray
        spikes = np.ones((2, 100))
        func(tmap, spikes, arena_size=1)
    with pytest.raises(err.ArgumentError):
        tmap = np.ones(10)
        spikes = ([0.23, 1], [0.5, 1.2], [0.75, -3])  #! spikes should be an ndarray
        func(tmap, spikes, arena_size=1)
    with pytest.raises(TypeError):
        tmap = np.ones((10, 10))
        spikes = np.ones((3, 100))  # t, x, y
        func(tmap, spikes)  #! missing arena_size
    with pytest.raises(err.DimensionMismatchError):
        tmap = np.ones((10, 10))
        spikes = np.ones((3, 100))
        func(tmap, spikes, arena_size=1, limits="abc")  # limits should be a tuple
    # mismatched sizes
    with pytest.raises(err.DimensionMismatchError):
        tmap = np.ones((40, 40))
        spikes = np.ones((3, 100))
        func(tmap, spikes, arena_size=80, bin_width=4)  #! 80/4 != 40

    print("test_rmap_invalid_inputs passed")
    return True


def test_1d_ratemap():
    arena_size = 80
    bin_width = 2.5
    _num_bins = opexebo.general.bin_width_to_bin_number(arena_size, bin_width)
    tmap = np.ones(_num_bins)  # equal occupancy in all locations
    n = 5000
    times = np.sort(np.random.rand(n)) * 1200  # 20 minute session
    speeds = np.random.rand(n) * 10  # linearly distributed speeds up to 10cm/s
    pos = np.random.rand(n) * arena_size
    spikes_tracking = np.array([times, speeds, pos])
    speed_cutoff = 2
    rmap = func(
        tmap,
        spikes_tracking,
        bin_width=bin_width,
        speed_cutoff=speed_cutoff,
        arena_size=arena_size,
    )
    assert rmap.ndim == 1
    assert rmap.shape == tmap.shape


def test_2d_ratemap():
    arena_size = (80, 120)
    bin_number = (16, 24)
    tmap = np.ones(
        bin_number
    ).T  # Want to be [x] 16 by [y] 24, whereas Numpy writes this the opposite, so transpose
    n = 5000
    times = np.sort(np.random.rand(n)) * 1200  # 20 minute session
    speeds = np.random.rand(n) * 10  # linearly distributed speeds up to 10cm/s
    pos_x = np.random.rand(n) * arena_size[0]
    pos_y = np.random.rand(n) * arena_size[1]
    spikes_tracking = np.array([times, speeds, pos_x, pos_y])
    speed_cutoff = 2
    rmap = func(
        tmap,
        spikes_tracking,
        bin_number=bin_number,
        speed_cutoff=speed_cutoff,
        arena_size=arena_size,
    )
    assert rmap.ndim == 2
    assert rmap.shape == tmap.shape


# if __name__ == '__main__':
#    test_rmap_invalid_inputs()
#    test_1d_ratemap()
#    test_2d_ratemap()



"""
Provide function for population vector correlation calculation
"""
import numpy as np

from .. import errors as err


def population_vector_correlation(stack_0, stack_1, **kwargs):
    """Calculates the bin-wise correlation between two stacks of rate maps
    
    Each stack corresponds to a separate Task, or trial. Each layer is the 
    ratemap for a single cell from that Task. The same units should be given in
    the same order in each stack.
    
    Take a single column through the stack (i.e. 1 single bin/location in
    arena, with a firing rate for each cell), from each stack
    
    In the original MatLab implementation, three output modes were supported
        * 1D: (`numYbins`) - iterate over `i`
            1) Take a 2D slice from each stack - all cells at all `X` positions at a
              single `Y` position `i`
            2) Reshape from 2D to 1D 
            3) Calculate the Pearson correlation coefficient between the two 1D
              arrays
            4) The value of `pv_corr_1d[i]` is the Pearson correlation coefficient
              arising from `Y` position `i`
        * 2D (`numXbins` x `numYbins`) - iterate over `i`
            1) Take a 2D slice from each stack - all cells at all `X` positions at a
              single `Y` position `i`
            2) Calculate the 2D array (`numXbins` x `numYbins`) where the `[j,k]`th
              value is the Pearson correlation coefficient between all
              observations at the `j`'th `X` location in `stack_left` and the `k`'th
              location in `stack_right`
            3) The `i`'th row of `pv_corr_2d` is the DIAGONAL of the correlation matrix
              i.e. where `j==k` i.e. the correlation of the the SAME location in
              each stack for all observations (`numCells`)
        * 3D (`numXbins` x `numYbins` x iteration(=`numYbins`))
            Same as 2D BUT take the whole correlation matrix, not the diagonal
            i.e. the full [j,k] correlatio between all X locations
    
    A note on correlation in Numpy vs Matlab
    
    Matlab's `corr(a, b)` function returns the correlation of ab
    Numpy's `corrcoef` function returns the normalised covariance matrix,
    which is:
            aa  ab
            ba  aa
    The normalised covariance matrix *should* be hermitian, but due to
    floating point accuracy, this is not actually guaranteed
    the MatLab function can be reproduced by taking either [0, 1] or [1,0]
    of the normalised covariance matrix. 

    If `a`, `b` are 2D matricies, then they should have shape `(num_variables, num_observations)`
    In the case of this function, where the iterator is over the `Y` values
    of the rate map, that means: `(x_bins, num_cells)`

    Parameters
    ----------
    stack_0: 3D array -or- list of 2D arrays
    stack_1: 3D array -or- list of 2D arrays
        `stack_x[i]` should return the `i`'th ratemap. This corresponds to a 
        constructor like:
            `np.zeros(num_layers, y_bins, x_bins)`
            
        Alternatively, a list or tuple of 2D arrays may be supplied:
            `stack_x` = (`ratemap_0`, `ratemap_1`, `ratemap_2`, ...)
    row_major: bool
        Direction of iteration. If `True`, then each row is iterated over in turn
        and correlation is calculated per row. 
        If `False`, then each column is iterated over in turn, and correlation is 
        calculated per column. 
        Default True (same behavior as in BNT)

    Returns
    -------
    (p1, p2, p3)
    p1: np.ndarray (1D, iterator x 1)
        Array of Pearson correlation coefficients. i'th value is given by the 
        correlation of the i'th flattened slice of stack_0 to the i'th
        flattened slice  of stack_1
    p2: np.ndarray (2D, iterator x non-iterator)
        i'th row is the diagonal of the correlation matrix, i.e. the correlation
        of the same location (location i) in each stack, i.e. where j==k
    p3: np.ndarray(3D, iterator x non-iterator x non-iterator)
        i'th array is the entire correlation matrix, rather than just the diagonal

    Notes
    --------
    BNT.+analyses.populationVectorCorrelation

    Copyright (C) 2019 by Simon Ball
    """
    debug = kwargs.get("debug", False)
    row_major = kwargs.get("row_major", True)
    
    # Perform input validation and ensure we have a pair of 3D arrays
    stack_0, stack_1 = _handle_both_inputs(stack_0, stack_1)
    
    # _handle_ has ensured that both arrays meet the shape/type requirements
    # Hardcode iterating over Y for now. 
    num_cells, y_bins, x_bins = stack_0.shape
    if row_major:
        iterator = y_bins
        non_iterator = x_bins
    else:
        iterator = x_bins
        non_iterator = y_bins
    
    if debug:
        print(f"Number of ratemaps: {num_cells}")
        print(f"Ratemap dimensions: {y_bins} x {x_bins}")
        print(f"Iterating over axis length {iterator} (row_major is {row_major})")

    p1 = np.zeros(iterator)
    p2 = np.zeros((iterator, non_iterator))
    p3 = np.zeros((iterator, non_iterator, non_iterator))

    for i in range(iterator):
        if row_major:
            left = stack_0[:, i, :].transpose()
            right = stack_1[:, i, :].transpose()
        else:
            left = stack_0[:, :, i].transpose()
            right = stack_1[:, :, i].transpose()
        
        # 1D
        # Reshape 2D array to a 1D array
        correlation_value = np.corrcoef(left.flatten(), right.flatten())[0,1]
        p1[i] = correlation_value
        
        # 2D, 3D
        correlation_matrix = np.corrcoef(left, right)[0:non_iterator, non_iterator:]
        p2[i, :] = np.diagonal(correlation_matrix)
        p3[i, :, :] = correlation_matrix

    return (p1, p2, p3)












###############################################################################
#############
#############           Error checking
#############


def _handle_both_inputs(stack_0, stack_1):
    '''Handle error checking across both main inputs'''
    stack_0 = _handle_single_input(stack_0, 0)
    stack_1 = _handle_single_input(stack_1, 1)
    if stack_0.shape[0] != stack_1.shape[0]:
        raise err.ArgumentError("You have a different number of rate maps in each stack.")
    if stack_0.shape[1:] != stack_1.shape[1:]:
        raise err.ArgumentError("Your rate maps do not have matching dimensions")
    return stack_0, stack_1

def _handle_single_input(stack, i):
    '''Handle the input stack(s) and provide a correctly formatted 3D array
    
    Handle error checking for a variety of conditions for a single stack
    If not already a MaskedArray, then convert to that
    
    Parameters
    ----------
    stack : array-like
        One of main inputs to population_vector_correlation.
        Should be either a 3D array, where each layer (stack[j]) is a RateMap,
        OR a list of 2D arrays, where each array is a 2D RateMap. 
        If a list of arrays, all arrays must be the same dimension
    i : int
        Index of stack input, solely used for providing more meaningful error
        message

    Returns
    -------
    stack : np.ma.MaskedArray
        3D array of RateMaps, masked at invalid values
    '''
    dims = None
    t = type(stack)
    if t not in (list, tuple, np.ndarray, np.ma.MaskedArray):
        raise ValueError(f"Stack_{i} must be array-like. You provided {t}")
    elif t in (tuple, list):
        for element in stack:
            e = type(element)
            if e not in (np.ndarray, np.ma.MaskedArray):
                raise err.ArgumentError(f"The elements of the list stack_{i} must be"\
                                 f" NumPy arrays. You provided {e}")
            if dims is None:
                dims = element.shape
            else:
                if element.shape != dims:
                    raise err.ArgumentError(f"Your ratemaps are not a consistent"\
                                     f" shape in stack_{i}")
        # Passes error handling, now convert from list to masked array
        stack = np.ma.masked_invalid(stack)
    elif isinstance(stack, np.ndarray):
        # Ok, but convert to masked array
        stack = np.ma.masked_invalid(stack)
        dims = stack.shape[1:]
    else:
        # Instance is already a Masked Array
        dims = stack.shape[1:]
    return stack


    


""" Tests for spatial occupancy"""
import numpy as np
import pytest
from opexebo.analysis import spatial_occupancy as func
from opexebo import errors

print("=== tests_analysis_spatial_occupancy ===")


def test_circular_arena():
    times = np.arange(5)
    positions = np.array([[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]]).T - 1
    speeds = np.ones(5)
    kwargs = {
        "arena_shape": "circ",
        "arena_size": 3,
        "bin_width": 1,
        "speed_cutoff": 0.1,
        "limits": (-2, 2.01, -2, 2.01),
    }
    map, coverage, bin_edges = func(times, positions, speeds, **kwargs)


#    import matplotlib.pyplot as plt
#    plt.imshow(map)
#    print(coverage, bin_edges)


def test_linear_arena():
    n = 1000
    times = np.arange(n)
    positions = np.random.rand(n)
    speeds = np.ones(n)
    kwargs = {
            "arena_shape": "line",
            "arena_size": 1,
            "bin_width": 0.1,
            "speed_cutoff": 0.1,
            "limits": (0, 1)
            }
    map, cov, edges = func(times, positions, speeds, **kwargs)
    assert cov == 1
    assert edges.size == 11
    assert edges[1] == 0.1
    
    # Check that it works with a 2d array of 1d position data
    positions = np.expand_dims(positions, 0)
    map, cov, edges = func(times, positions, speeds, **kwargs)
    assert cov == 1
    assert edges.size == 11
    assert edges[1] == 0.1

def test_invalid_inputs():
    # wrong dimensions to positions
    n = 10

    with pytest.raises(errors.ArgumentError):
        times = np.arange(n)
        positions = np.ones((3, n))  #!
        speeds = np.ones(n)
        func(times, positions, speeds, arena_size=1)
    with pytest.raises(errors.ArgumentError):
        times = np.arange(n)
        positions = np.ones((2, n))
        speeds = np.ones((2, n))  #!
        func(times, positions, speeds, arena_size=1)

    # Mismatched pos/speed
    with pytest.raises(errors.ArgumentError):
        times = np.arange(n)
        positions = np.ones((2, n))
        speeds = np.ones(n + 1)  #!
        func(times, positions, speeds, arena_size=1)

    # No arena size
    with pytest.raises(TypeError):
        times = np.arange(n)
        positions = np.ones((2, n))
        speeds = np.ones(n)
        func(times, positions, speeds)  #!

    # All nan
    # This isn't explicit in the function, but comes as a result of excluding
    # all non-finite values, and then being left with an empty array
    with pytest.raises(ValueError):
        times = np.arange(n)
        positions = np.full((2, n), np.nan)
        speeds = np.full(n, np.nan)
        func(times, positions, speeds, arena_size=1)
    
    with pytest.raises(NotImplementedError):
        times = np.arange(5)
        positions = np.ones(5)
        speeds = np.ones(5)
        func(times, positions, speeds, arena_size=1, arena_shape="asdf")

    print("test_invalid_inputs passed")


#if __name__ == '__main__':
#    test_circular_arena()
#    test_linear_arena()
#    test_invalid_inputs()




'''Opexebo default values

This module provides default values for all keyword arguments in the Opexebo 
package. 

'''


'''spatial/angular occupancy'''
#: Standard spatial resolution, [cm]
bin_width = 2.5

#: Standard angular bin width [degree]
bin_angle = 15

#: Standard speed bin, [cm/s]
bin_speed = 5

#: Standard speed cutoff below which to ignore positions [cm/s]
speed_cutoff = 0

#: Shape of area: assume a rectangular arena
shape = "rectangular"
shapes_square = ("square", "rectangle", "rectangular", "rect", "s", "r")
shapes_circle = ("circ", "circular", "circle", "c")
shapes_linear = ("linear", "line", "l")



'''Smoothing'''
#: Standard Gaussian Standard Deviation for smoothing, [bins]
sigma = 2.0

#: Standard Gaussian stdev for smoothing firing rat, [s]. Default at 0.4s from doi 10.1038/nature14622 .     
sigma_time = 0.4

#: Replacement value for masked values when smoothing. Use np.nan to interpolate through masked values instead of use fixed value
mask_fill = 0



'''LFP related'''
#: Default method for calculating a power spectrum
power_spectrum_method = "welch"

#: Default effective resolution for Welch's method (Hz)
psd_resolution_welch = 1

#: Default return scale, i.e. linear or decibel scale
psd_return_scale = "log"



''' Firing field related'''
#: Initial relative threshold to search for local maxima, in range [0,1]
initial_search_threshold = 0.96

#: Minimum number of bins for a firing field to be valid [bins]
firing_field_min_bins = 9

#: Minimum peak firing rate for a firing field to be valid [Hz]
firing_field_min_peak = 1

#: Minimum mean firing rate for a firing field to be valid [Hz]
firing_field_min_mean = 0.1

#: The method used to find local maxima
search_method = 'default'

#: All implemented means of finding local maxima. Use lower case. 
all_methods = (search_method, "sep", "not implemented")    




''' Score related'''
#: Percentile for head drection arc, in range [0,1]
hd_percentile = 0.95

#: minimum angular separation between fields in acorr considered for grid ellipse [degrees]
min_orientation = 15

#: The default speed bandwidth for adaptive bandpassing [cm/s]
speed_bandwidth = 2
#: The default lower edge of the bandpass filter used for speedscore [cm/s]
lower_bound_speed = 2
#: The default time used to calculate the upper edge of the bandpass filter [s]
upper_bound_time = 10
#: The default upper edge of the bandpass filter if a fixed (rather than adaptive) filter is chosen [cm/s]
upper_bound_speed = 15


#: Search width for border coverage, see analyses.bordercoverage, [bins]
search_width = 8
#: standard walls for border search, see analyses.bordercoverage, string    
walls = 'trbl'


""" Tests for get_speed"""
import numpy as np
import math
import pytest
from opexebo.analysis import calc_speed as func


def test_invalid_inputs():
    n = 1000
    t = np.arange(n)
    invalid_kwargs = [
        {"t": t, "x": np.arange(n - 1)},
        {"t": t, "x": t, "y": np.arange(n - 1)},
        {"t": t, "x": 3},
        {"t": t, "x": t, "moving_average": 2.3},
        {"t": t, "x": t, "moving_average": 2},
        {"t": t, "x": t, "moving_average": 4},
    ]
    for kwargs in invalid_kwargs:
        with pytest.raises(ValueError):
            func(**kwargs)
    return


def test_valid_inputs():
    n = 1000
    t = np.arange(n, dtype=np.float64)
    nan_x = t.copy()
    nan_x[27] = np.nan
    inf_x = t.copy()
    inf_x[35] = np.inf
    mass_nan_x = t.copy()
    mass_nan_x[np.random.rand(n) > 0.6] = np.nan
    valid_kwargs = [
        {"t": t, "x": t},
        {"t": t, "x": t, "y": t},
        {"t": t, "x": t, "moving_average": 5},
        {"t": t, "x": nan_x},
        {"t": t, "x": inf_x},
        {"t": t, "x": mass_nan_x},
    ]
    for kwargs in valid_kwargs:
        func(**kwargs)
    return


def test_known_cases():
    # constant speed:
    n = 1000
    t = np.arange(n)
    x = np.arange(n)
    speed = func(t, x)
    assert math.isclose(np.std(speed), 0, abs_tol=1e-6)
    assert speed[0] == 1
    assert speed[-1] == 1

    # constant acceleration
    # Should yield a speed that is equal to arange(n) EXCEPT for speed[0]
    t = np.arange(n)
    x = np.zeros(n)
    for i in np.arange(n - 1):
        x[i + 1] = x[i] + i + 1
    speed = func(t, x)
    accel = np.diff(speed[1:])
    assert math.isclose(np.std(accel), 0, abs_tol=1e-6)
    assert accel[0] == 1
    assert speed[0] == 1
    assert speed[57] == 57

    # Orbiting a point
    t = np.arange(n)
    r = 28
    x = r * np.cos(2 * np.pi * t / n)
    y = r * np.sin(2 * np.pi * t / n)
    speed = func(t, x, y)
    assert math.isclose(np.std(speed), 0, abs_tol=1e-6)
    d = 2 * r * np.pi
    assert math.isclose(speed[193], d / n, rel_tol=1e-4)



import numpy as np
import skimage.transform


def upsample(array, upscale, **kwargs):
    '''A MaskedArray aware upsampling routine. 
    
    `skimage` provides more reliable methods for interpolated upsampling of simple
    ndarrays. However, this is not appropriate to MaskedArrays, i.e. where
    certain locations have invalid values that should not expand.
    
    For non-masked arrays, this function can optionally rely on the `skimage`
    routine allowing fractional upscaling. For MaskedArrays, integer upscaling
    is enforced, with no interpolation
    
    Parameters
    ----------
    array: array-like
        Supports `list`, `tuple`, `np.ndarray` or `np.ma.MaskedArray`. Supports
        1d and 2d (2d for true arrays only). If the array contains any non-finite
        values, or is a MaskedArray, then integer upscaling is enforced
    upscale: int or float
        Upscaling factor. The output array shape will be this factor larger.
        Method of upscaling depends on type(upscale). 
            * `int` -> integer upscaling with no interpolation
            * `float` -> upscaling with interpolation
        Note the type distinction: `type(1) == int`; `type(1.) == float`
    debug: bool, optional
        Print out debugging information while running. Default False

    Returns
    -------
    new_array : np.ndarray or np.ma.MaskedArray
        Upsampled array
    '''
    ### General handling of the input array
    # If a tuple or list was provided, convert to an ndarray
    if isinstance(array, (tuple, list)):
        array = np.array(array)
    # Check for invalid values
    if isinstance(array, np.ndarray):
        if np.logical_not(np.isfinite(array)).any():
            # Check for any non-finite elements -> convert to masked array if so
            array = np.ma.masked_invalid(array)
        else:
            pass
    elif isinstance(array, np.ma.MaskedArray):
        pass
    else:
        raise NotImplementedError(f"Upsampling is not supported for array type {type(array)}")

    debug = kwargs.get("debug", False)
    
    if debug:
        print(f"Array type: {type(array)}")
        print(f"Upscaling: {upscale}, type {type(upscale)}")
    
    #####           Based on the inputs, decide how to proceed
    # Is the array Masked (then use integer scaling), or not masked (then allow non-integer scaling
    if isinstance(array, np.ma.MaskedArray):
        if isinstance(upscale, int):
            new_array = _integer_upsampling(array, upscale, **kwargs)
        else:
            raise NotImplementedError(f"Fractional interpolation is not supported"\
                                      f" with MaskedArrays. You requested upscaling {upscale}")
    else:
        if isinstance(upscale, int):
            new_array = _integer_upsampling(array, upscale, **kwargs)
        else:
            new_array = _fractional_upsampling(array, upscale, **kwargs)
    
    return new_array

###############################################################################
####            Actual calculation functions here

    
def _integer_upsampling(array, upscale, **kwargs):
    '''
    Perform integer upscaling on the provided array. No interpolation

    Each pixel in `array` is replaced by `upscale`x`upscale` pixels of the same
    value in `new_array`.

    Parameters
    ----------
    array: np.ndarray or np.ma.MaskedArray
        1D or 2D, the array to be scaled
    upscale: int
        upscaling factor, >= 1. 1 is an acceptable, though pointless, value -
        the original array is returned with no modifications

    Returns
    -------
    new_array : np.ndarray or np.ma.MaskedArray
        Same array type as input. 
    '''
    if upscale < 1:
        raise NotImplementedError("upscale must be 1 or higher")
    elif upscale == 1:
        new_array = array
    elif isinstance(array, np.ma.MaskedArray):
        #In the case of a MaskedArray, upscale the mask and data separately
        ndata = _integer_upsampling(array.data, upscale, **kwargs)
        nmask = _integer_upsampling(np.ma.getmaskarray(array), upscale, **kwargs)
        new_array = np.ma.masked_where(nmask, ndata)
    else:
        new_shape = np.array(array.shape) * upscale
        new_array = np.zeros(new_shape, dtype=array.dtype)

        # 1D
        if array.ndim == 1:
            for i in range(upscale):
                new_array[i:new_shape[0]+i+1-upscale:upscale] = array[:]

        # 2D
        elif array.ndim == 2:
            for i in range(upscale):
                for j in range(upscale):
                    new_array[i:new_shape[0]+i+1-upscale:upscale,
                              j:new_shape[1]+j+1-upscale:upscale] = array[:,:]
        else:
            raise NotImplementedError(f"Integer upscaling only supports 1D and 2D arrays."\
                                      f" You provided {array.ndim}D")
    return new_array


def _fractional_upsampling(array, upscale, **kwargs):
    '''Perform fractional upscaling on the provided array. Since there is no
    direct mapping from one to many locations, use interpolation to calculate
    the new values
    
    Parameters
    ----------
    array: np.ndarray
        nD array to be scaled
    upscale: float
        Scaling factor. Must be > 0
    
    Returns
    -------
    new_array : np.ndarray
        Upscaled, interpolated array. Same type and dimensionality as input array
    '''
    if upscale == 1:
        new_array = array
    elif upscale <= 0:
        raise ValueError(f"Upscaling not supported for values <= 0 ({upscale})")
    elif not isinstance(array, np.ndarray):
        raise ValueError(f"Upscaling is only supported for np.ndarrays. You provided {type(array)}")
    else:
        new_array = skimage.transform.rescale(array, upscale)
    return new_array


# Basic functions
from .speed import calc_speed

# 2D spatial functions : functional
from .spatial_occupancy import spatial_occupancy
from .rate_map import rate_map
from .rate_map_stats import rate_map_stats
from .rate_map_coherence import rate_map_coherence
from .place_field import place_field
from .autocorrelation import autocorrelation
from .grid_score import grid_score, grid_score_stats
from .egocentric_occupancy import egocentric_occupancy

# 1D angular functions : functional
from .angular_occupancy import angular_occupancy
from .tuning_curve import tuning_curve
from .tuning_curve_stats import tuning_curve_stats

# Miscellanious
from .population_vector_correlation import population_vector_correlation
from .theta_modulation_index import theta_modulation_index

# Experimental
from .speed_score import speed_score
from .border_coverage import border_coverage
from .border_score import border_score

__all__ = ["calc_speed", 
        "spatial_occupancy", "rate_map", "rate_map_stats", "rate_map_coherence",
        "grid_score", "grid_score_stats", "autocorrelation", "place_field", 
        "egocentric_occupancy",
           "angular_occupancy", "tuning_curve", "tuning_curve_stats", 
           "population_vector_correlation", "theta_modulation_index",
           "border_coverage", "border_score", "speed_score"]



""" Provides functions for calculating Speed Score """

import numpy as np
import opexebo.defaults as default
import opexebo
from opexebo.errors import SpeedBandwidthError, ArgumentError


def speed_score(spike_times, tracking_times, tracking_speeds, **kwargs):
    '''
    Calculate Speed score.

    Speed score is a correlation between cell firing *rate* and animal speed.
    The Python version is based on BNT.+scripts.speedScore. At Edvard's request,
    both the 2015 and 2016 scores are calculated. The primary difference is how
    the speed smoothing is implemented

    Speed score originates in the following paper in Nature from Emiliano et al
    doi:10.1038/nature14622

    The original Matlab script implemented - but as far as I can tell, did not
    (by default) use, a Kalman filter for smoothing the animal speed. Since it is
    not the default behaviour, I have not (yet) added that Kalman filter to opexebo.
    Its addition is contingent on the score similarity to BNT.

    Summary:
        * The intention is to correlate (spike firing rate) with (animal speed)
        * Convert an N-length array of spike firing times into an M-length array
          of spike firing rates, where M is the same length as tracking times
        * Optional: smooth firing rates
        * Optional: smooth speeds
        * Optional: apply a bandpass filter to speeds
        * calculate the Pearson correlation coefficient between (speed), (firing rate)

    Parameters
    ----------
    spike_times: np.ndarray
        N-length array listing the times at which spikes occurred. [s]

    tracking_times: np.ndarray
        M-length array of time stamps of tracking frames

    tracking_speeds: np.ndarray
        M-length array of animal speeds at time stamps given in `tracking_times`

    Other Parameters
    ----------------
    bandpass: str
        Type of bandpass filter applied to animal speeds. Acceptable values
        are
            * `"none"` - No speed based filtering is applied
            * `"fixed"` - a fixed lower and upper speed bound are used, based
              on keywords `"lower_bound_speed"`, `"upper_bound_speed"`
            * `"adaptive"` - a fixed lower speed bound is used, based on
              keyword `"lower_bound_speed"`. An upper speed bound is determined
              based on keywords `"upper_bound_time"` and "`speed_bandwidth"`
        Default `none`
    lower_bound_speed: float
        Speed in [cm/s] used as the lower edge of the speed bandpass filter
        (`"fixed"` and `"adaptive"`). Default 2cm/s
    upper_bound_speed: float
        Speed in [cm/s] used as the upper edge of the speed bandpass filter
        (`"fixed"` only). Default 15cm/s
    upper_bound_time: float
        Duration in [s] used for determining the upper edge of the speed
        bandpass filter (`"adaptive"` only). Default 10s
    speed_bandwidth: float
        Range of speeds in [cm/s] used for determining the upper edge of the
        speed bandpass filter (`"adaptive"` only). Default 2cm/s
    sigma: float
        Standard deviation in [s] of Gaussian smoothing kernel for smoothing
        both speed and firing rate data. Default 0.5s
    debug: bool

    Returns
    -------
    scores: dict
        2015: float
        2016: float
            Variations on the speed score. '2015' is based on the code in the paper
            above, but additionally including an upper speed filter
            '2016' is a modification involving a slightly different approach to
            smoothing the firing rate data
    (lower_speed, upper_speed): list of floats
        Speed thresholds using in the bandpass filter
        Most useful in the case of the adaptive filter, because there is no
        other way to find out what was actually used.

    Notes
    -----
    BNT.+scripts.speedScore

    Copyright (C) 2019 by Simon Ball
    '''
    # Check that the provided arrays have correct dimensions
    if spike_times.ndim != 1:
        raise ArgumentError("spike_times must be an Nx1 array. You have provided"\
                         f" {spike_times.ndim} dimensions")
    elif tracking_times.ndim != 1:
        raise ArgumentError("tracking_times must be an Nx1 array. You have provided"\
                         f" {tracking_times.ndim} dimensions")
    elif tracking_speeds.ndim != 1:
        raise ArgumentError("tracking_speeds must be an Nx1 array. You have provided"\
                         f" {tracking_speeds.ndim} dimensions")
    if tracking_times.size != tracking_speeds.size:
        raise ArgumentError("tracking_times and tracking_speeds must be the same length")

    # Get kwargs values
    speed_bandwidth = kwargs.get('speed_bandwidth', default.speed_bandwidth)
    sigma_time = kwargs.get('sigma', default.sigma_time)
    upper_bound_time = kwargs.get('upper_bound_time', default.upper_bound_time) # Only used in "adaptive" filter
    lower_bound_speed = kwargs.get('lower_bound_speed', default.lower_bound_speed)
    upper_bound_speed = kwargs.get("upper_bound_speed", default.upper_bound_speed) # Only used in "fixed" filter
    bandpass_type = kwargs.get("bandpass", "none").lower()
    available_filters = ("none", "fixed", "adaptive")
    if bandpass_type not in available_filters:
        raise NotImplementedError(f"Bandpass tpye '{bandpass_type}' is not implemented."\
                                  f" Available types are {available_filters}.")
    debug = kwargs.get('debug', False)

    # Convert spike_times to spike firing rate
    sampling_rate = 1 / np.mean(np.diff(tracking_times))
    firing_rate = _spiketimes_to_spikerate(spike_times, tracking_times, sampling_rate)

    # Apply smoothing
    # Smoothing expects to be given a sigma in units [bins], so convert from real units to bins
    tracking_speeds_smoothed = opexebo.general.smooth(tracking_speeds, sigma_time * sampling_rate)
    firing_rate_smoothed = opexebo.general.smooth(firing_rate, sigma_time * sampling_rate)

    # Calculate the bandpass filter
    if debug:
        print(bandpass_type)
    if bandpass_type == "none":
        _filter = _bandpass_none(tracking_speeds_smoothed, **kwargs)
        lower_bound_speed = 0
        upper_bound_speed = np.inf
    elif bandpass_type == "fixed":
        _filter = _bandpass_fixed(tracking_speeds_smoothed, lower_bound_speed,
                                  upper_bound_speed, **kwargs)
    elif bandpass_type == "adaptive":
        _filter, upper_bound_speed = _bandpass_adaptive(tracking_speeds_smoothed, sampling_rate,
                                                        lower_bound_speed, upper_bound_time, speed_bandwidth, **kwargs)

    # Apply the filter to speeds
    speeds = tracking_speeds_smoothed[_filter]
        # The filter will be applied differently to rate based on which score version is wanted

    # Score 2016: apply bandpass filter to already-smoothed rate and then correlate
    rate = firing_rate_smoothed[_filter]
    speed_score_2016 = np.corrcoef(speeds, rate)[0, 1]

    # Score 2015: Filter rates first (by setting to NaN), and then smooth and correlate
    # Reuse the same filtered_speeds as for 2016, but redefine filtered_rate
    rate = firing_rate[_filter]
    rate = opexebo.general.smooth(rate, sigma_time * sampling_rate)
    speed_score_2015 = np.corrcoef(speeds, rate)[0, 1]

    scores = {'2015': speed_score_2015, '2016': speed_score_2016}
    return scores, (lower_bound_speed, upper_bound_speed)






def _bandpass_adaptive(speed, sampling_rate, lower_speed, upper_time, speed_bw, **kwargs):
    '''Create a filter list that allows through values based on a defined lower
    value, and an upper value determined by the highest 2cm/s bandwidth at which
    the animal spent at least X time

    Calculating the upper speed:
        * Histogram the speed array with a resolution 10x higher than the
        desired speed-bandwidth
        * Iterate over the resulting histogram to identify the highest speed range
        at which the animal spends at least upper_time
        * select the centre of this speed range as the upper threshold

    parameters
    ----------
    speed : np.ndarray
        1d M-length array of animal speeds at fixed sample rate, in [cm/s]
    sampling_rate : float
        Sampling rate of the tracking system, in [Hz]
    lower_speed : float
        Lower threshold of bandpass filter, in [cm/s]
    upper_time : float
        Time for calculating upper_speed in [s]. The upper_speed is calculated
        as the highest [2cm/s] speed bandwidth that the animal spends at least
        this long.
    speed_bw : float
        Range of speeds for calculating the upper_speed, in [cm/s]

    returns
    -------
    _filter : np.ndarray
        1d M-length array of booleans for indexing the speed array. True where
        the speed PASSES the filter
    '''
    multiplier = 10
    hist_resolution = speed_bw / multiplier          # this is bin_width
    bins = np.arange(np.min(speed), np.max(speed), hist_resolution)
    hist, _ = np.histogram(speed, bins=bins)

    required_frames = upper_time * sampling_rate
    upper_speed = None

    for i in np.arange(-1, -(bins.size - multiplier), -1):
        # Iterate backwards through the histogram, i.e. from highest speeds
        total_frames = np.sum(hist[i:i+multiplier])
        if total_frames >= required_frames:
            # go to the centre of the bandwidth. minus because of reverse direction
            upper_speed = bins[i-int(multiplier/2)]
            break
    if kwargs.get("debug", False):
        print(f"Upper speed determined as {upper_speed} cm/s")
    if upper_speed is not None:
        _filter = _bandpass_fixed(speed, lower_speed, upper_speed, **kwargs)
    else:
        raise SpeedBandwidthError(f"The animal did not spend {upper_time}s within a speed"\
                         f" bandwidth of {speed_bw} cm/s. Try using a"\
                         " larger speed-bandwidth")
    return _filter, upper_speed

def _bandpass_fixed(speed, lower_speed, upper_speed, **kwargs):
    '''Create a filter list that allows through values between the defined upper
    and lower defined values

    parameters
    ----------
    speed : np.ndarray
        1d M-length array of animal speeds at fixed sample rate, in [cm/s]
    sampling_rate : float
        Sampling rate of the tracking system, in [Hz]
    lower_speed : float
        Lower threshold of bandpass filter, in [cm/s]
    upper_speed : float
        Upper threshold of bandpass filter, in [cm/s]
        Required upper_speed > lower_speed

    returns
    -------
    _filter : np.ndarray
        1d M-length array of booleans for indexing the speed array. True where
        the speed PASSES the filter
    '''
    if lower_speed >= upper_speed:
        raise ValueError(f"Your lower bound ({lower_speed}) is higher than your"\
                         f" upper bound ({upper_speed}). Check your argument order.")
    _filter = (lower_speed <= speed) & (speed <= upper_speed)
    passed = np.sum(_filter)
    if kwargs.get("debug", False):
        print(f"{passed:,} survived filter out of {speed.size:,} ({passed/speed.size:3})")
    if passed <= 5:
        raise ValueError("Your filter has excluded nearly all values, only"\
                         f" {passed} remaining. Check your filter criteria")
    return _filter

def _bandpass_none(speed, **kwargs):
    '''Create a filter list that allows all values through'''
    debug = kwargs.get("debug", False)
    if debug:
        print("No filtering")
    _filter = np.ones(speed.size, dtype=bool)
    return _filter

def _spiketimes_to_spikerate(spike_times, tracking_times, sampling_rate):
    '''Convert a list of spike times to a list of spike rates

    parameters
    ----------
    spike_times : np.ndarray
        Nx1 array of times at which spikes occur in [s]
    tracking times : np.ndarray
        Nx1 array of times at which tracking information is known - e.g. time
        stamp of camera frames. Also in [s]

    returns
    -------
    spike_rate : np.adarray
        Nx1 array of spike rate [Hz], with the i'th value being the spike rate
        during the i'th tracking frame.
    '''
    frame_length = 1/sampling_rate
    bin_edges = np.append(tracking_times, tracking_times[-1]+frame_length)

    spikes_per_frame, _ = np.histogram(spike_times, bins=bin_edges)
    spike_rate = spikes_per_frame *sampling_rate

    return spike_rate



"""Provide function for calculating statistics of a Rate Map"""

import numpy as np


def rate_map_stats(rate_map, time_map, debug=False):
    '''
    Calculate statistics of a rate map that depend on probability distribution
    function (PDF)
    
    Calculates information, sparsity and selectivity of a rate map. Calculations
    are done according to 1993 Skaggs et al. "An Information-Theoretic Approach
    to Deciphering the Hippocampal Code" paper. Another source of information is
    1996 Skaggs et al. paper called "Theta phase precession in hippocampal
    neuronal populations and the compression of temporal sequences".

    Coherence is calculated based on RU Muller, JL Kubie "The firing of
    hippocampal place cells predicts the future position of freely moving rats",
    Journal of Neuroscience, 1 December 1989, 9(12):4101-4110. The paper doesn't
    provide information about how to deal with border values which do not have
    8 well-defined neighbours. This function uses zero-padding technique.

    Parameters
    ----------
    rate_map: np.ma.MaskedArray
        Smoothed rate map: n x m array where cell value is the firing rate,
        masked at locations with low occupancy

    time_map: np.ma.MaskedArray
        time map: n x m array where the cell value is the time the animal spent
        in each cell, masked at locations with low occupancy
        Already smoothed

    Returns
    -------
    rms: dict
        spatial_information_rate: float
            information rate [bits/sec]
        spatial_information_content: float
            spatial information content [bits/spike]
        sparsity: float
            see relevant literature (above)
        selectivity: float
            see relevant literature (above)
        peak_rate: float
            peak firing rate of smoothed map [Hz]
        mean_rate: float
            mean firing rate of smoothed map [Hz]

    Notes
    -----
    BNT.+analyses.mapStatsPDF(map)
    
    BNT.+analyses.coherence(map)

    Copyright (C) 2019 by Simon Ball
    '''

    if type(rate_map) != np.ma.MaskedArray:
        rate_map = np.ma.masked_invalid(rate_map, copy=True)
    if type(time_map) != np.ma.MaskedArray:
        time_map = np.ma.masked_invalid(time_map, copy=True)

    duration = np.ma.sum(time_map)
    position_PDF = time_map / (duration + np.spacing(1)) 
    # Probability distribution of where the animal spent its time. 

    if debug:
        print("Duration = %ds" % duration)
        print("Masked locations: %d" % np.sum(rate_map.mask))
        print("Masked values: %s" % rate_map.data[rate_map.mask])

    sparsity = np.nan
    selectivity = np.nan
    inf_rate = np.nan
    inf_content = np.nan
    if rate_map.mask.all():
        # Currently, there is a bug in numpy that causes nanmean to fail
        # on fully masked arrays. This should be a pretty rare occurrence, though!
        rmap_mean = np.nan
        rmap_peak = np.nan
    else:
        rmap_mean = np.nanmean(rate_map)
        rmap_peak = np.nanmax(rate_map)

    mean_rate = np.ma.sum( rate_map * position_PDF )
    mean_rate_sq = np.ma.sum( np.ma.power(rate_map, 2) * position_PDF )

    max_rate = np.max(rate_map)

    if debug:
        print("mean rate: %.2fHz" % mean_rate)
        print("mean rate squared: %.2fHz^2" % mean_rate_sq)
        print("max rate: %.2fHz" % max_rate)


    if mean_rate_sq != 0:
        sparsity = mean_rate * mean_rate / mean_rate_sq

    if mean_rate != 0:
        selectivity = max_rate / mean_rate

        log_argument = rate_map / mean_rate
        log_argument[log_argument < 1] = 1
        if debug:
            print(log_argument.shape)
            #print("log argument: %.4f" % log_argument)
        inf_rate = np.ma.sum(position_PDF * rate_map * np.ma.log2(log_argument))
        inf_content = inf_rate / mean_rate



    return {"spatial_information_rate":inf_rate, "spatial_information_content":inf_content,
            "sparsity":sparsity, "selectivity":selectivity, "peak_rate":rmap_peak,
            "mean_rate":rmap_mean}



"""Interpret the 'arena_size' keyword"""

import numpy as np

import opexebo.defaults as default
import opexebo.errors as err

def validatekeyword__arena_size(kwv, provided_dimensions):
    '''
    Decipher the possible meanings of the keyword "arena_size".
    
    "arena_size" is given to describe the arena in which the animal is moving
    It should be either a float, or an array-like of 2 floats (x, y)
    
    Parameters
    ----------
    kw: float or array-like of floats
        The value given for the keyword `arena_size`
    provided_dimensions : int
        the number of spatial dimensions provided to the original function.
        Acceptable values are 1 or 2
        E.g. if the original function was provided with positions = [t, x, y], then
        provided_dimensions=2 (x and y)

    Returns
    -------
    arena_size : float or np.ndarray of floats
    
    Raises
    ------
    ValueError
    IndexError
    '''
    if provided_dimensions == 1:
        is_2d = False
    elif provided_dimensions == 2:
        is_2d = True
    else:
        raise NotImplementedError("Only 1d and 2d arenas are supported. You"\
                                  " provided %dd" % provided_dimensions)
    if type(kwv) in (float, int, str):
        kwv = float(kwv)
        if kwv <= 0: 
            raise err.ArgumentError("Keyword 'arena_size' value must be greater than"\
                             " zero (value given %f)" % kwv)
        if is_2d:
            arena_size = np.array((kwv, kwv))
        else:
            arena_size = kwv
    elif type(kwv) in (list, tuple, np.ndarray):
        if len(kwv) == 1:
            if is_2d:
                
                arena_size = np.array(kwv[0], kwv[0])
            else:
                arena_size = kwv[0]
        elif len(kwv) == 2 and not is_2d:
            raise err.DimensionMismatchError("Mismatch in dimensions: 1d position data but 2d"\
                             " arena specified")
        elif len(kwv) not in [1, 2]:
            raise err.ArgumentError("Keyword 'arena_size' value is invalid. Provide"\
                             " either a float or a 2-element tuple")
        else:
            arena_size = np.array(kwv)
    else:
        raise err.ArgumentError("Keyword 'arena_size' value not understood. Please"\
                         " provide either a float or a tuple of 2 floats. Value"\
                         " provided: '%s'" % str(kwv))
    return arena_size, is_2d


def validate_keyword_arena_shape(arena_shape):
    '''
    Ensure that the arena_shape is a meaningful value
    
    Parameters
    ----------
    arena_shape : str
        the value given for the keyword `arena_shape`
    
    Returns
    -------
    arena_shape : str
        A value that is guaranteed to be an acceptable member of one of the
        recognised groups of arena_shapes
    '''
    if not isinstance(arena_shape, str):
        raise err.ArgumentError("Keyword `arena_shape` must be a string, not type `{type(arena_shape)}`")
    else:
        arena_shape = arena_shape.lower()
    
    if arena_shape in default.shapes_square:
        # this is ok
        pass
    elif arena_shape in default.shapes_circle:
        # this is ok
        pass
    elif arena_shape in default.shapes_linear:
        # this is ok
        pass
    else:
        raise NotImplementedError(f"Arena shape '{arena_shape}' not implemented")
    
    return arena_shape
    
    


'''
Provides function to calculate the angular occupancy, i.e. the frequency with 
which the animal looked in each direction
'''
import numpy as np
import opexebo
import opexebo.defaults as default
import opexebo.errors as errors


def angular_occupancy(time, angle, **kwargs):
    '''
    Calculate angular occupancy from tracking angle and kwargs over (0,2*pi)

    Parameters
    ----------
    time : numpy.ndarray
        time stamps of angles in seconds
    angle : numpy array
        Head angle in radians
        Nx1 array
    bin_width : float, optional
        Width of histogram bin in degrees

    Returns
    -------
    masked_histogram : numpy masked array
        Angular histogram, masked at angles at which the animal was never 
        observed. A mask value of True means that the animal never occupied
        that angle. 
    coverage : float
        Fraction of the bins that the animal visited. In range [0, 1]
    bin_edges : list-like
        x, or (x, y), where x, y are 1d np.ndarrays
        Here x, y correspond to the output histogram
    
    Notes
    --------
    Copyright (C) 2019 by Simon Ball, Horst Obenhaus

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    '''
    if time.ndim != 1:
        raise errors.ArgumentError("time must be provided as a 1D array. You provided %d"\
                         " dimensions" % time.ndim)
    if angle.ndim != 1:
        raise errors.ArgumentError("angle must be provided as a 1D array. You provided %d"\
                         " dimensions" % angle.ndim)
    if time.size != angle.size:
        raise errors.ArgumentError("Arrays 'time' and 'angle' must have the same number"\
                         f" of elements. You provided {time.size} and {angle.size}")
    if time.size == 0:
        raise errors.ArgumentError("Zero length array provided when data expected")
    if np.nanmax(angle) > 2*np.pi:
        raise Warning("Angles greater than 2pi detected. Please check that your"\
                      " angle array is in radians. If it is in degrees, you can"\
                      " convert with 'np.radians(array)'")

    bin_width = kwargs.get('bin_width', default.bin_angle)
    bin_width = np.radians(bin_width)
    arena_size = 2*np.pi
    limits = (0, arena_size)

    angle_histogram, bin_edges = opexebo.general.accumulate_spatial(angle, bin_width=bin_width, 
                                                arena_size=arena_size, limits=limits)
    masked_angle_histogram = np.ma.masked_where(angle_histogram==0, angle_histogram)
    
    # masked_angle_histogram is in units of frames. It needs to be converted to units of seconds
    frame_duration = np.mean(np.diff(time))
    masked_angle_seconds = masked_angle_histogram * frame_duration
    
    # Calculate the fractional coverage based on locations where the histogram
    # is zero. If all locations are  non-zero, then coverage is 1.0
    coverage = np.count_nonzero(angle_histogram) / masked_angle_seconds.size
    
    return masked_angle_seconds, coverage, bin_edges



"""Provide a function for mapping a list of positional data into a 1D or 2D space"""

import numpy as np
from opexebo.general import validatekeyword__arena_size, bin_width_to_bin_number
import opexebo.defaults as default


def accumulate_spatial(pos, arena_size, **kwargs):
    """
    Given a list of positions, create a histogram of those positions. The
    resulting histogram is typically referred to as a map.
    
    The complexity in this function comes down to selecting where the edges of
    the arena are, and generating the bins within those limits.
    
    The histogram bin edges must be defined in one of 3 different ways
    
        * bin_width: based on the keyword `arena_size`, the number of bins will
          be calculated as
            `opexebo.general.bin_width_to_bin_number`
          The histogram will use `num_bins` between the minimum and maximum of the
          positions (or `limit` if provided)
        * bin_number: the histogram will use bin_number of bins between the
          minimum and maximum of the positions (or `limit` if provided)
        * bin_edges: the histogram will use the provided `bin_edg`e arrays

    Either zero or one of the three bin_* keyword arguments must be defined.
    If none are defined, then a default bin_width is used. If more than 1 is
    defined, an error is raised

    Parameters
    ----------
    pos: np.ndarray
        1D or 2D array of positions  in row-major format, i.e. `x` = pos[0],
        `y` = pos[1]. This matches the simplest input creation pos = np.array( [`x`, `y`] )
    arena_size: float or tuple of floats
        Dimensions of arena (in cm)
            * For a linear track, length
            * For a circular arena, diameter
            * For a rectangular arena, length or (length, length)
    bin_width: float
        Bin size in cm. Default 2.5cm. If bin_width is supplied, `limit` must
        also be supplied. One of `bin_width`, `bin_number`, `bin_edges` must be
        provided
    bin_number: int or tuple of int
        Number of bins. If provided as a tuple, then `(x_bins, y_bins)`. One
        of `bin_width`, `bin_number`, `bin_edges` must be provided
    bin_edges: array-like
        Edges of the bins. Provided either as `edges` or `(x_edges, y_edges)`.
        One of `bin_width`, `bin_number`, `bin_edges` must be provided
    limits: tuple or np.ndarray
        (x_min, x_max) or (x_min, x_max, y_min, y_max)
        Provide concrete limits to the range over which the histogram searches
        Any observations outside these limits are discarded
        If no limits are provided, then use np.nanmin(data), np.nanmax(data)
        to generate default limits.
        As is standard in python, acceptable values include the lower bound
        and exclude the upper bound

    Returns
    -------
    hist: np.ndarray
        1D or 2D histogram of the occurrences of the input observations
        Dimensions given by arena_size/bin_width
        Not normalised - each cell gives the integer number of occurrences of
        the observation in that cell
    edges: list-like
        `x`, or (`x`, `y`), where `x`, `y` are 1d np.ndarrays
        Here `x`, `y` correspond to the output histogram
    """

    # Check correct inputs:
    pos = np.squeeze(pos.copy()) 
    dims = pos.ndim
    if dims not in (1, 2):
        raise ValueError("pos should have either 1 or 2 dimensions. You have"\
                         " provided %d dimensions." % dims)

    # Get kwargs values
    debug = kwargs.get("debug", False)
    limits = kwargs.get("limits", None)
    if not isinstance(limits, (tuple, list, np.ndarray, type(None))):
        raise ValueError("You must provide an array-like 'limits' value, e.g."\
          " (x_min, x_max, y_min, y_max). You provided type %s" % type(limits))

    arena_size, is_2d = validatekeyword__arena_size(arena_size, dims)

    
    ###########################################################################
    ####### Handle the decision of bin_edges
    # Logic:
        # If none are provided, use default bin_width
        # If more than 1 is provided, raise Exception
        # If bin_edges are provided, use them
        # Else use bin_width, if provided
        # Else use bin_number, if provided
    bin_number = kwargs.get("bin_number", None)
    bin_width = kwargs.get("bin_width", None)
    bin_edges = kwargs.get("bin_edges", None)
    if bin_edges is None and bin_width is None and bin_number is None:
        # No bin decision was provided, so go with default
        bin_width = default.bin_width
        debug_bin_type = "default - bin_width"
    elif sum( [x is not None for x in (bin_edges, bin_width, bin_number)] ) != 1:
        # Count the number of values where a value other than False is present
        # If there are more than 1 "True" value, then the user has provided too many keywords
        raise KeyError("You have provided more than one method for determining"\
                       " the edges of the histogram. Only zero or one methods"\
                       " can be accepted.")

    if bin_edges is not None:
        # First priority: use predefined bin_edges
        # Remember - user provides (x, y), but histogram needs (y, x)
        if is_2d:
            if not isinstance(bin_edges, (tuple, list, np.ndarray)):
                raise ValueError("keyword 'bin_edges' must be either a tuple or list"\
                                 " (of np.ndarrays), or a 2D array")
        else:
            if not isinstance(bin_edges, np.ndarray):
                raise ValueError("Keyword 'bin_edges' must be a numpy array for a 1D histogram")
        bins = (bin_edges[1], bin_edges[0])
        debug_bin_type = "bin_edges"

    elif bool(bin_width):
        # Calculate the number of bins based on the requested width and arena_size
        # Then calculate the actual bin edges that this would give, based on expanding from top left
        # Given arena size in (x, y), this also returns in (x, y) -> have to convert for Numpy
        num_bins = bin_width_to_bin_number(arena_size, bin_width)
        if limits is None:
            # Handle the case that limits is not provided, i.e. is None
            lim = (0, 0, 0, 0)
        else:
            lim = limits
        if is_2d:
            # Have to swap to (y, x)
            bins = (np.linspace(0, arena_size[1], num_bins[1]+1) + lim[2],
                    np.linspace(0, arena_size[0], num_bins[0]+1) + lim[0])
        else:
            bins = np.linspace(0, arena_size, num_bins+1) + (lim[0])
        debug_bin_type = "bin_width"

    elif bool(bin_number):
        if isinstance(bin_number, int):
            bins = bin_number
        elif isinstance(bin_number, (tuple, list, np.ndarray)):
            # If an array, we expect to recieve (x, y)
            # Have to convert for the Numpy standard of (y, x)
            bins = (bin_number[1], bin_number[0])
        else:
            raise ValueError("Keyword 'bin_number' must be an integer, or an"\
                             " array-like of integers.")
        debug_bin_type = "bin_number"

    if debug:
        print(f"Limits: {limits}")
        print(f"Binning type: {debug_bin_type}")
        print(f"bins : {bins}")

    ###########################################################################
    ###### Make the histogram
    if is_2d:
        x = pos[0]
        y = pos[1]
        if limits is None:
            limits = np.array([np.nanmin(y), np.nanmax(y)*1.001, np.nanmin(x),
                               np.nanmax(x)*1.001]).reshape(2, 2)
            # numpy convention: (y, x)
            if debug:
                print("No limits found. Calculating based on min/max")
        elif len(limits) != 4:
            raise ValueError("You must provide a 4-element 'limits' value for a"\
                             " 2D map. You provided %d elements" % len(limits))
        else:
            limits = np.flipud(np.array(limits).reshape(2, 2))
            # the flipud swaps x and y - rememebr, numpy convention that (y, x)
        in_range = np.logical_and(np.logical_and(np.greater_equal(y, limits[0, 0]), 
                                                 np.less(y, limits[0, 1])),
                                  np.logical_and(np.greater_equal(x, limits[1, 0]), 
                                                 np.less(x, limits[1, 1]))
                                  )
        # the simple operator ">= doesn't respect Masked Arrays
        # As of 2019, it does actually behave correctly (NaN is invalid and so
        # is removed), but I would prefer to be explicit

        in_range_x = x[in_range]
        in_range_y = y[in_range]
        if debug:
            print(f"data points : {len(in_range_x)}")

        '''
        A brief word on the documentation for np.histogram2d()
        https://docs.scipy.org/doc/numpy/reference/generated/numpy.histogram2d.html
        
        The documentation is subtlely misleading in the use of `x` and `y`. 
        
        The NumPy standard notation is (almost) invariably to call (y, x), and, e.g.,
        in a 2D array, you would create an array with 5 rows and 2 columns like follows:
            np.zeros((5, 2))
        
        We would normally describe this as an array with a height (corresponding to y)
        of 5, and a width (corresponding to x) of 2.
        
        The documentation for histogram2d gives the signature as:
            numpy.histogram2d(x, y, bins=10, range=None, normed=None, weights=None, density=None)
        
            Returns:
                H : ndarray, shape(nx, ny)
                    The bi-dimensional histogram of samples x and y. Values in x are 
                    histogrammed along the first dimension and values in y are histogrammed 
                    along the second dimension.
                xedges : ndarray, shape(nx+1,)
                    The bin edges along the first dimension.
                yedges : ndarray, shape(ny+1,)
                    The bin edges along the second dimension.
        
        Note the order in the Returns section: data given as `x` is histogrammed along the
        _first_ dimension, which in NumPy parlance, would _usually_ be labelled `y`.
        The use of `x`, `y` is self-consistent within this page, but misleading
        in the context of other NumPy functions. 
        
        By invoking `y` as the first axis, and then returning edges as [xedges, yedges],
        we are internally consistent with the mathematical notation that I have used throughout
        opexebo (i.e. to call x, then y)
        '''
        hist, yedges, xedges = np.histogram2d(in_range_y, in_range_x, bins=bins, range=limits)
        edges = [xedges, yedges]

    else: # is not 2d
        x = pos
        if limits is None:
            limits = [np.nanmin(x), np.nanmax(x)*1.0001]
        elif len(limits) != 2: 
            raise ValueError("You must provide a 2-element 'limits' value for a"\
                             " 1D map. You provided %d elements" % len(limits))
        in_range = np.logical_and(np.greater_equal(x, limits[0]),
                                  np.less(x, limits[1]))
        in_range_x = x[in_range]
        if debug:
            print(f"data points : {len(in_range_x)}")

        hist, edges = np.histogram(in_range_x, bins=bins, range=limits)

    return hist, edges



""" Tests for autocorrelogram"""
import numpy as np
import inspect

from opexebo.analysis import rate_map_stats as func

print("=== tests_rate_map_stats ===")


###############################################################################
################                MAIN TESTS
###############################################################################


def test_random_data():
    """Doesn't test that meaningful results are produced, only that something is"""
    n = 40
    tmap = np.random.rand(n, n) * 5
    rmap = np.random.rand(n, n)
    func(tmap, rmap)
    print(f"{inspect.stack()[0][3]} passed")
    return True


def test_zero_data():
    """Doesn't test that meaningful results are produced, only that something is"""
    n = 40
    tmap = np.ma.zeros((n, n))
    rmap = np.ma.zeros((n, n))
    func(tmap, rmap)
    print(f"{inspect.stack()[0][3]} passed")
    return True


# if __name__ == '__main__':
#    test_random_data()
#    test_zero_data()
